
import abc, weakref, uuid, traceback as tb
import sys
import threading

from .types import *

from ...Misc import mixedmethod
from ..JHTML import JHTML, DefaultOutputWidget
from ..JHTML.WidgetTools import JupyterAPIs, frozendict

class JHTMLConversionError(Exception):
    """
    Represents an error in converting to JHTML
    """
    def __init__(self, widgets, cause, tb, message="failed to convert to JHTML:\n{}"):
        self.widgets = widgets
        self.base_cause = cause
        self.base_tb = tb
        self.message_template = message
        super().__init__(self.format_message())
    def format_message(self, limit=10):
        widget_chain = "\n".join(("  >> " if i > 0 else "") + repr(w) for i,w in enumerate(self.widgets))
        cause = "\n".join(tb.format_exception(None, self.base_cause, self.base_tb, limit=-limit))
        return self.message_template.format(widget_chain + "\n" + cause)

__all__ = [
    "WidgetInterface",
    "GenericDisplay",
    "DelayedResult",
    "Component",
    "WrapperComponent",
    "Container",
    "MenuComponent",
    "ListGroup",
    "Button",
    "LinkButton",
    "Spinner",
    "Progress",
    "ButtonGroup",
    "Navbar",
    "Carousel",
    "Pagination",
    "Sidebar",
    "Dropdown",
    "DropdownList",
    "Tabs",
    "TabPane",
    "TabList",
    "Accordion",
    "AccordionHeader",
    "AccordionBody",
    "Opener",
    "OpenerHeader",
    "OpenerBody",
    "CardOpener",
    "Modal",
    "ModalHeader",
    "ModalBody",
    "ModalFooter",
    "Offcanvas",
    "OffcanvasHeader",
    "OffcanvasBody",
    "Toast",
    "ToastBody",
    "ToastHeader",
    "ToastContainer",
    "Spacer",
    "Breadcrumb",
    "Card",
    "CardHeader",
    "CardBody",
    "CardFooter",
    "ModifierComponent",
    "Tooltip",
    "Popover",
    "Layout",
    "Grid",
    "Table",
    "Flex"
]
__reload_hook__ = ["..JHTML", "..WidgetTools"]

class WidgetInterface(metaclass=abc.ABCMeta):
    """
    Provides the absolute minimum necessary for hooking
    an interface that creates an `ipywidget` into the
    Jupyter display runtime
    """
    @abc.abstractmethod
    def to_widget(self):
        ...
    # @abc.abstractmethod
    def initialize(self):
        ...
    def _ipython_display_(self):
        JupyterAPIs.get_display_api().display(self.to_widget())
        self.initialize()

    _display_locks = set()
    def display(self):
        if self not in self._display_locks: # don't want to call this over and over...
            self._display_locks.add(self)
            try:
                self._ipython_display_()
            finally:
                self._display_locks.remove(self)
    def get_mime_bundle(self):
        return self.to_widget().get_mime_bundle()
    @mixedmethod
    def _ipython_pinfo_(self):
        from ...Docs import jdoc
        return jdoc(self)

class Component(WidgetInterface):
    """
    Provides an abstract base class for an interface element
    to allow for the easy construction of interesting interfaces
    """
    def __init__(self, dynamic=True, debug_pane=None, **attrs):
        super().__init__()
        self._parents = weakref.WeakSet()
        self._widget_cache = None
        attrs['dynamic'] = dynamic
        self._attrs = attrs
        self.debug_pane = DefaultOutputWidget.get_default() if debug_pane is None else debug_pane
    @property
    def attrs(self):
        return frozendict(self._attrs)
    @attrs.setter
    def attrs(self, value):
        self._attrs = value

    def get_attr(self, key):
        return self._attrs[key]
    def get_child(self, key):
        raise NotImplementedError("{} doesn't have children".format(
            type(self).__name__
        ))
    def __getitem__(self, item):
        if isinstance(item, str):
            return self.get_attr(item)
        else:
            return self.get_child(item)

    def set_attr(self, key, value):
        self._attrs[key] = value
    def update_widget_attr(self, key, value):
        self._widget_cache[key] = value
    def set_child(self, which, new):
        raise NotImplementedError("{} doesn't have children".format(
            type(self).__name__
        ))
    def update_widget_child(self, key, value):
        self._widget_cache[key] = value
    def __setitem__(self, key, value):
        if isinstance(key, str):
            self.set_attr(key, value)
            if self._widget_cache is not None:
                self.update_widget_attr(key, value)
        else:
            self.set_child(key, value)
            if self._widget_cache is not None:
                self.update_widget_child(key, value)

    def del_attr(self, key,):
        del self._attrs[key]
    def del_widget_attr(self, key):
        del self._widget_cache[key]
    def del_child(self, which):
        raise NotImplementedError("{} doesn't have children".format(
            type(self).__name__
        ))
    def del_widget_child(self, key):
        del self._widget_cache[key]
    def __delitem__(self, key):
        if isinstance(key, str):
            self.del_attr(key)
            if self._widget_cache is not None:
                self.del_widget_attr(key)
        else:
            self.del_child(key)
            if self._widget_cache is not None:
                self.del_widget_child(key)

    def insert(self, where, new):
        self.insert_child(where, new)
        if self._widget_cache is not None:
            self.insert_widget_child(where, new)
    def append(self, child):
        return self.insert(None, child)

    def insert_child(self, where, child):
        raise NotImplementedError("{} doesn't have children".format(
            type(self).__name__
        ))
    def insert_widget_child(self, where, child):
        self._widget_cache.insert(where, child)

    def add_class(self, *cls):
        self.add_component_class(*cls)
        if self._widget_cache is not None:
            self.add_widget_class(*cls)
    def add_component_class(self, *cls):
        if not 'cls' in self._attrs:
            self._attrs['cls'] = []
        new = self._attrs['cls'].copy()
        for c in cls:
            for y in JHTML.manage_class(c):
                if y not in new:
                    new.append(y)
        self._attrs['cls'] = new
    def add_widget_class(self, *cls):
        return self._widget_cache.add_class(*cls)
    def remove_class(self, *cls):
        self.remove_component_class(*cls)
        if self._widget_cache is not None:
            self.remove_widget_class(*cls)
    def remove_component_class(self, *cls):
        if not 'cls' in self._attrs:
            self._attrs['cls'] = []
        new = self._attrs['cls'].copy()
        for c in cls:
            for y in JHTML.manage_class(c):
                try:
                    new.remove(y)
                except ValueError:
                    pass
        self._attrs['cls'] = new
    def remove_widget_class(self, *cls):
        return self._widget_cache.remove_class(*cls)

    @abc.abstractmethod
    def to_jhtml(self, parent=None):
        ...
    def to_widget(self, parent=None):
        if parent is not None:
            self._parents.add(parent)
        if self._widget_cache is None:
            with DefaultOutputWidget(self.debug_pane):
                try:
                    self._widget_cache = self.to_jhtml()
                except JHTMLConversionError as e:
                    raise JHTMLConversionError(e.widgets + [self], e.base_cause, e.base_tb) from None
                except:
                    _, e, tb = sys.exc_info()
                    raise JHTMLConversionError([self], e, tb) from None
                self._widget_cache.component = self
            # self._widget_cache.to_widget.observe(self.set_value, )
        return self._widget_cache
    def mutate(self, fn):
        fn(self)
        self.invalidate_cache()
    def invalidate_cache(self):
        self._widget_cache = None
        for w in self._parents:
            w.invalidate_cache()
class WrapperComponent(Component):
    """
    Extends the base component interface to allow for the
    construction of interesting compound interfaces (using `JHTML.Compound`).
    Takes a `dict` of `wrappers` naming the successive levels of the interface
    along with a `theme` that provides style declarations for each level.

    Used primarily to create `Bootstrap`-based interfaces.
    """
    wrappers = dict(wrapper=JHTML.Div)
    theme = dict(wrapper={'cls': []})
    def __init__(self,
                 items: ElementType,
                 wrappers=None,
                 theme=None,
                 extend_base_theme=True,
                 **attrs):

        self.items, attrs = self.manage_items(items, attrs)

        if wrappers is None:
            wrappers = self.wrappers
        self.wrappers = wrappers

        wrappers = list(wrappers.items())

        self.theme = self.manage_theme(theme, extend_base_theme=extend_base_theme)
        self.theme[wrappers[0][0]] = self.merge_themes(self.theme.get(wrappers[0][0], {}), attrs)
        if len(wrappers) > 1:
            attrs = {'wrapper_attrs':self.theme}
            self.wrapper = JHTML.Compound(*[
                (key, wrapper)
                for i, (key, wrapper)
                in enumerate(wrappers)
            ])
        else:
            attrs = self.theme[wrappers[0][0]]
            self.wrapper = wrappers[0][1]
        super().__init__(**attrs) # need to delegate attr updates to the theme...

        # self.item_attrs = theme.get(item[0], {})
        # self.item = item[-1]
    def handle_variants(self, theme):
        if 'variant' in theme or 'base-cls' in theme:
            theme = theme.copy()
            cls = theme.get('cls', [])
            if isinstance(cls, str):
                cls = cls.split()
            theme['cls'] = list(cls) + [theme.get('base-cls', cls[0] if len(cls) > 0 else "")+"-"+theme.get('variant', '')]
            try:
                del theme['variant']
            except KeyError:
                pass
            try:
                del theme['base-cls']
            except KeyError:
                pass
        return theme
    def manage_theme(self, theme, extend_base_theme=True):
        if theme is None:
            theme = self.theme
        theme = theme.copy()
        if extend_base_theme:
            for k,v in self.theme.items():
                if k in theme:
                    theme[k] = self.merge_themes(v, theme[k])
                else:
                    theme[k] = self.theme[k].copy()
        return theme
    @classmethod
    def merge_themes(cls, theme: 'None|dict', attrs:dict, merge_keys=('cls',)):
        """
        Needs to handle cases where a `theme` is provided
        which includes things like `cls` declarations and then the
        `attrs` may also include `cls` declarations and the `attrs`
        declarations get appended to the theme
        """

        if theme is None:
            theme = {}
        theme = theme.copy()

        kinter = theme.keys() & attrs.keys()
        if merge_keys is not None:
            kinter = kinter & set(merge_keys)

        for k in attrs:
            if k in kinter:
                if isinstance(theme[k], str):
                    theme[k] = theme[k].split()
                if isinstance(attrs[k], str):
                    attrs[k] = attrs[k].split()
                if attrs[k] is None:
                    attrs[k] = []
                if isinstance(theme[k], dict):
                    theme[k] = cls.merge_themes(theme[k], attrs[k], merge_keys=None)
                else:
                    theme[k] = theme[k] + attrs[k]
            else:
                theme[k] = attrs[k]

        return theme

    @classmethod
    def manage_items(cls, items, attrs):
        if isinstance(items, dict):
            attrs = dict(attrs, **items)
            del attrs['body']
            items = items['body']
        elif (
                isinstance(items, tuple)
                and len(items) == 2
                and isinstance(items[1], dict)
        ):
            attrs = dict(attrs, **items[1])
            items = items[0]
        if (
                isinstance(items, (str, int, float, JupyterAPIs.get_widgets_api().Widget))
                or hasattr(items, 'to_tree')
                or hasattr(items, 'to_widget')
        ):
            items = [items]
        elif items is None:
            items = []
        else:
            items = list(items)
        return items, attrs
    def get_child(self, key):
        return self.items[key]
    def set_child(self, which, new):
        self.items[which] = new
    def insert_child(self, where, child):
        if where is None:
            where = len(self.items)
        self.items.insert(where, child)
    # def add_component_class(self, *cls):
    #     if 'cls' not in self.attrs:
    #         base_cls = None
    #     self.add
    #         self.attrs['cls'] =
    #     new = self.wrapper_classes.copy()
    #     for c in cls:
    #         for y in JHTML.manage_class(c):
    #             if y not in new:
    #                 new.append(y)
    #     self.wrapper_classes = new
    # def remove_component_class(self, *cls):
    #     new = self.wrapper_classes.copy()
    #     for c in cls:
    #         for y in JHTML.manage_class(c):
    #             try:
    #                 new.remove(y)
    #             except ValueError:
    #                 pass
    #     self.wrapper_classes = new

    def wrap_items(self, items):
        if isinstance(self.wrappers, JHTML.Compound):
            attrs = self.attrs.copy()
            wrapper_attrs = self.attrs.get('wrapper_attrs', {}).copy()
            for k,v in wrapper_attrs:
                wrapper_attrs[k] = self.handle_variants(v.copy())
            del attrs['wrapper_attrs']
            return self.wrapper(*items, wrapper_attrs=wrapper_attrs, **attrs)
        else:
            return self.wrapper(*items, **self.handle_variants(self.attrs))
    def to_jhtml(self, parent=None):
        return self.wrap_items(self.items)
class Container(WrapperComponent):
    """
    Extends the base `WrapperComponent` to include a final
    `items` spec for cases where there is a base wrapper and a set of items,
    e.g. a list group which has the `list-group` outer class and a set of `list-items` inside.
    """
    wrappers = dict(wrapper=JHTML.Div, item=JHTML.Span)
    theme = dict(wrapper={'cls':[]}, item={'cls':[]})
    def __init__(self,
                 items: ElementType,
                 wrappers: dict = None,
                 **attrs) -> None:

        if wrappers is None:
            wrappers = self.wrappers
        self.wrappers = wrappers

        wrappers = list(wrappers.items())
        if len(wrappers) == 1:
            wrappers.append([JHTML.Span, {}])
        item = wrappers[-1]
        wrappers = dict(wrappers[:-1])


        self._items = None
        items, attrs = self.manage_items(items, attrs)
        super().__init__(None, wrappers=wrappers, **attrs)
        self._items = items

        self.item_attrs = self.theme.get(item[0], {}).copy()
        self.item = item[-1]

        if isinstance(self.item, dict): # a way to specify a subtheme
            item_wrappers = list(self.item.items())
            item_theme = self.item_attrs
            if len(item_wrappers) > 1:
                self.item_attrs = {k[0]:item_theme.get(k[0], {}) for k in item_wrappers}
                self.item = JHTML.Compound(*[
                    (key, wrapper)
                    for i, (key, wrapper)
                    in enumerate(item_wrappers)
                ])
            else:
                self.item_attrs = item_theme.get(wrappers[0][1], {})
                self.item = item_wrappers[0][1]

    @property
    def items(self):
        return [self.create_item(i) for i in self._items]
    @items.setter
    def items(self, items):
        if self._items is not None:
            raise ValueError("can't set items")
    def _create_dict_item(self, body=None, **extra):
        if isinstance(self.item, JHTML.Compound):
            wrapper_attrs = self.item_attrs.copy()
            for k, v in wrapper_attrs.items():
                wrapper_attrs[k] = self.handle_variants(v)
            n, _ = self.item.destructure_wrapper(self.item.base) # get base name
            if n is not None:
                wrapper_attrs[n] = self.merge_themes(wrapper_attrs.get(n, {}), extra)
            return self.item(body, wrapper_attrs=wrapper_attrs)
        else:
            return self.item(body, **self.merge_themes(self.handle_variants(self.item_attrs), extra))
    def _create_base_item(self, body):
        if isinstance(self.item, JHTML.Compound):
            return self.item(body, wrapper_attrs=self.item_attrs)
        else:
            return self.item(body, **self.item_attrs)
    def create_item(self, i, **kw):
        if isinstance(i, dict):
            if 'raw' in i:
                return i['raw']
            if len(kw) > 0:
                return self._create_dict_item(**dict(i, **kw))
            else:
                return self._create_dict_item(**i)
        else:
            if len(kw) > 0:
                return self._create_dict_item(body=i, **kw)
            else:
               return self._create_base_item(i)

    def update_widget_child(self, key, value):
        super().update_widget_child(key, self.create_item(value))
    def insert_widget_child(self, where, child):
        super().insert_widget_child(where, self.create_item(child))
class ComponentContainer(WrapperComponent):
    components = {}
    def __init__(self, component_args:dict=None, component_kwargs:dict=None,components=None, **attrs):
        super().__init__([], **attrs)
        if components is None:
            components = {}
        self.components = dict(self.components, **components)
        self.component_args = component_args if component_args is not None else {}
        self.component_kwargs = component_kwargs if component_kwargs is not None else {}
    def create_components(self):
        return {
            k:c(
                *self.component_args.get(k, []),
                theme=self.theme.get(k, {}),
                **self.component_kwargs.get(k, {})
            )
            for k,c in self.components.items()
            if not (len(self.component_args.get(k, [])) == 1 and len(self.component_kwargs.get(k, [])) == 0 and self.component_args.get(k, [])[0] is None) # components to ignore
        }
    def handle_variants(self, theme):
        return theme
    def wrap_items(self, items):
        items = list(self.create_components().values())
        return super().wrap_items(items)
class ModifierComponent(Component):
    modifiers = None
    def __init__(self, base=None, **modifiers):
        base_mods = self.modifiers
        if base_mods is None:
            base_mods = {}
        modifiers = dict(base_mods, **modifiers)
        super().__init__(**modifiers)
        self.base = base
    def __call__(self, base):
        if self.base is None:
            self.base = base
        else:
            raise ValueError("{} already has a base object".format(
                self
            ))
        return self
    blacklist = {'dynamic'}
    def to_jhtml(self):
        base = self.base
        if hasattr(base, 'to_jhtml'):
            base = base.to_jhtml()
        for k,v in self.attrs.items():
            if k not in self.blacklist:
                base[k] = v
        return base

class Button(WrapperComponent):
    wrappers = dict(button=JHTML.Bootstrap.Button)
    theme = dict(button={'cls':['btn'], 'variant':'primary'})
    def __init__(self, body, action=None, event_handlers=None, **kwargs):
        if event_handlers is None:
            event_handlers = {}
        self._action = action
        if isinstance(action, (str, dict, list)):
            event_handlers['click'] = action
        else:
            event_handlers['click'] = self._eval
        self._eval_lock = None
        super().__init__(
            body,
            event_handlers=event_handlers,
            **kwargs
        )
    @property
    def action(self):
        return self._action
    @action.setter
    def action(self, a):
        self._action = a
        if isinstance(a, (str, dict, list)):
            self._attrs['event_handlers'] = {'click':a}
            if self._widget_cache is not None:
                self._widget_cache.add_event(click=a)
        else:
            self._attrs['event_handlers'] = {'click':self._eval}
            if self._widget_cache is not None:
                self._widget_cache.add_event(click=self._eval)
    def _eval(self, *args):
        if self.action is not None and self._eval_lock is None:
            self._eval_lock = True
            w = self.to_widget()
            try:
                pane = w.debug_pane
            except AttributeError:
                pane = None
            try:
                if pane is None:
                    self.action(*args)
                else:
                    with pane:
                        self.action(*args)
            finally:
                self._eval_lock = None
class LinkButton(Button):
    wrappers = dict(button=JHTML.Anchor)
    theme = dict(button={'cls':[]})

class Spinner(WrapperComponent):
    wrappers = {'spinner':JHTML.Div}
    theme = {'spinner':{'cls':[], 'base-cls':'spinner', 'variant':'border'}}
    def __init__(self, body=None, role='status', **attrs):
        """
        :param variant:
        :type variant:
        :param attrs:
        :type attrs:
        """
        if body is None:
            body = JHTML.Span("Loading...", cls="visually-hidden")
        super().__init__(body, role=role, **attrs)

class Progress(WrapperComponent):
    wrappers = {'wrapper':JHTML.Div, 'bar':None}
    theme = {'wrapper':{'cls':['progress']}, 'bar':{'cls':['progress-bar']}}
    def __init__(self, value=0, label=None, wrappers=None, **attrs):
        items = []
        if label is True:
            label = str(value) + "%"
        elif not label:
            label = None
        if label is not None:
            items.append(label)
        if wrappers is None:
            wrappers = self.wrappers.copy()
        if 'bar' not in wrappers:
            wrappers['bar'] = None
        if wrappers['bar'] is None:
            wrappers['bar'] = JHTML.Styled(JHTML.Div, width=str(value)+"%", dynamic=True)
        super().__init__(
            items,
            wrappers=wrappers,
            **attrs
        )
    @property
    def container(self):
        return self.to_widget().compound_wrapper_data["container"]
    @property
    def bar(self):
        return self.to_widget().compound_wrapper_data["bar"]
    def update_widget_attr(self, attr, val):
        self.container[attr] = val

#region Menus
class MenuComponent(Container):
    def __init__(self, items:ElementType, **attrs):
        super().__init__(items, **attrs)
        self._item_map = {}
    def create_item(self, item, **kw):
        item = super().create_item(item, **kw)
        if hasattr(item, 'id'):
            self._item_map[item.id] = item
        elif 'id' in item.attrs:
            self._item_map[item.attrs['id']] = item
        return item
class ListGroup(MenuComponent):
    theme = {
        'wrapper':{'cls':['list-group', 'list-group-flush']},
        'item':{'cls':['list-group-item', 'list-group-item-action']}
    }
class ButtonGroup(MenuComponent):
    wrappers = {
        'wrapper':JHTML.Div,
        'item':Button
    }
    theme = {
        'wrapper':{'cls':['btn-group']}
    }
class DropdownList(MenuComponent):
    wrappers = {
        'wrapper': JHTML.Ul,
        'item': {'list-item':JHTML.Li, 'link':JHTML.Anchor}
    }
    theme = {
        'wrapper': {'cls': ['dropdown-menu']},
        'item': {'list-item':{}, 'link':{'cls': ['dropdown-item']}},
    }
    def create_item(self, item, **kw):
        if isinstance(item, dict) and 'action' in item:
            item = item.copy()
            item['event_handlers'] = {'click':item['action']}
            del item['action']
        item = super().create_item(item, **kw)
        return item
class Dropdown(ComponentContainer):
    components = {
        'toggle':Button,
        'list':DropdownList
    }
    theme = {
        'wrapper':{'cls':['dropdown']},
        'toggle':{'button':{'cls':['dropdown-toggle'], 'data-bs-toggle':'dropdown'}}
    }
    def __init__(self, header:ElementType, actions:ElementType, **attrs):
        super().__init__(
            {
                'toggle':(header,),
                'list':(self.prep_actions(actions),)
            },
            **attrs
        )
    def prep_actions(self, actions):
        if isinstance(actions, dict):
            acts = []
            for k,v in actions.items():
                if isinstance(v, tuple) and len(v) == 2 and isinstance(v[1], dict):
                    v, opts = v
                else:
                    opts = {}
                acts.append(dict(opts, body=k, action=v))
            actions = acts
        return actions

class Navbar(MenuComponent):
    wrappers = {
        'wrapper':JHTML.Nav,
        'container':JHTML.Div,
        'nav':JHTML.Div,
        'item':JHTML.Anchor
    }
    theme = {
        'wrapper': {'cls':['navbar', 'navbar-expand']},
        'container': {'cls':['container-fluid']},
        'nav': {'cls':['navbar-nav']},
        'item': {'cls':['nav-link']}
    }
class Sidebar(MenuComponent):
    wrappers = {
        'wrapper': JHTML.Nav,
        'item': {'list-item':JHTML.Li, 'link':JHTML.Anchor}
    }
    theme = {
        'wrapper': {'cls': ['nav', 'flex-column']},
        'item': {
            'list-item': {'cls': ['nav-item']},
            'link': {'cls': ['nav-link']}
        }
    }

class Pagination(MenuComponent):
    wrappers = {
        'wrapper': JHTML.Nav,
        'container': JHTML.Ul,
        'item': {'page-item': JHTML.Li, 'link': JHTML.Anchor}
    }
    theme = {
        'wrapper': {'cls': []},
        'container': {'cls': ['pagination']},
        'item': {
            'page-item': {'cls': ['page-item']},
            'link': {'cls': ['page-link']}
        }
    }


def short_uuid(len=6):
    return str(uuid.uuid4()).replace("-", "")[:len]

class Carousel(MenuComponent):
    wrappers = {
        'wrapper': JHTML.Div,
        'inner': JHTML.Div,
        'item': JHTML.Div
    }
    theme = {
        'wrapper': {'cls': ['carousel']},
        'inner': {'cls': ['carousel-inner']},
        'item': {'cls': ['carousel-item']}
    }
    def __init__(self, items, include_controls=True, data_bs_ride='carousel', interval=None, **attrs):
        self.include_controls = include_controls
        self._active_made = False
        self.base_name = 'carousel-' + short_uuid()
        self.interval = interval
        super().__init__(items, id=self.base_name, data_bs_ride=data_bs_ride, **attrs)
    def create_item(self, item, cls=None, data_bs_interval=None, **kw):
        cls = JHTML.manage_class(cls)
        if not self._active_made:
            cls = cls + ['active']
            self._active_made = True
        if data_bs_interval is None:
            data_bs_interval = str(self.interval) if self.interval is not None else "10000000000"
        return super().create_item(item, cls=cls, data_bs_interval=data_bs_interval, **kw)
    def next_button(self, body=None, cls='carousel-control-next', **kwargs):
        return JHTML.Button(
                    JHTML.Span(cls='carousel-control-next-icon') if body is None else body,
                    **self.merge_themes(dict(cls=cls, data_bs_target='#'+self.base_name, data_bs_slide='next'), kwargs)
                )
    def prev_button(self, body=None, cls='carousel-control-prev', **kwargs):
        return JHTML.Button(
            JHTML.Span(cls='carousel-control-prev-icon') if body is None else body,
            **self.merge_themes(
                dict(cls=cls, data_bs_target='#' + self.base_name, data_bs_slide='prev'),
                kwargs
            )
        )
    def wrap_items(self, items):
        base = super().wrap_items(items)
        if self.include_controls:
            base.append(self.prev_button())
            base.append(self.next_button())
        return base


class TabList(MenuComponent):
    wrappers = {
        'wrapper': JHTML.Ul,
        'item': {'list-item':JHTML.Li, 'tab-button':JHTML.Button}
    }
    theme = {
        'wrapper': {'cls':['nav'], 'variant':'tabs'},
        'item': {'list-item': {'cls':['nav-item']}, 'tab-button': {'cls':['nav-link']}}
    }
    def __init__(self, *args, base_name=None, role="tablist", **kwargs):
        if base_name is None:
            base_name = 'tabs-' + str(uuid.uuid1()).replace("-", "")[:5]
        self.base_name = base_name
        self._active = None
        super().__init__(*args, role=role, **kwargs)
    def create_item(self, item, cls=None, **kw):
        # cls = JHTML.manage_class(cls)
        # if not self._active_made:
        #     cls = cls + ['active']
        #     self._active_made = True
        # if data_bs_interval is None:
        #     data_bs_interval = str(self.interval) if self.interval is not None else "10000000000"
        # return super().create_item(item, cls=cls, data_bs_interval=data_bs_interval, **kw)
        cls = JHTML.manage_class(cls)
        item, _ = item
        item_id = self.base_name+"-"+item.replace(' ', '')
        if self._active is None:
            cls = cls + ['active']
            self._active = item_id
        return super().create_item(item, id=item_id+'-tab', cls=cls, data_bs_target='#'+item_id, data_bs_toggle='tab', **kw)
class TabPane(MenuComponent):
    wrappers = {
        'wrapper': JHTML.Div,
        'item': JHTML.Div
    }
    theme = {
        'wrapper': {'cls': ['tab-content']},
        'item': {'cls': ['tab-pane']}
    }
    def __init__(self, *args, base_name=None, **kwargs):
        if base_name is None:
            base_name = 'tabs-' + short_uuid()
        self.base_name = base_name
        self._active = None
        super().__init__(*args, **kwargs)
    def create_item(self, item, cls=None, **kw):
        key, item = item
        item_id = self.base_name+"-"+key.replace(' ', '')
        cls = JHTML.manage_class(cls)
        if self._active is None:
            cls = cls + ['active']
            self._active = item_id
        item = super().create_item(item, id=item_id, role='tabpanel', cls=cls, **kw)
        return item
class Tabs(ComponentContainer):
    components = {
        'list':TabList,
        'pane':TabPane
    }
    theme = {
        'pane':{},
        'list': {}
    }
    def __init__(self, tabs, base_name=None, **attrs):
        if base_name is None:
            base_name = 'tabs-' + str(uuid.uuid1()).replace("-", "")[:5]
        if isinstance(tabs, dict):
            tabs = tabs.items()
        super().__init__(
            {
                'list': (tabs,),
                'pane': (tabs,)
            },
            {
                'list':dict(id=base_name, base_name=base_name),
                'pane':dict(id=base_name, base_name=base_name)
            },
            **attrs
        )

class AccordionHeader(Container):
    wrapper_classes = ['accordion-header']
    item = JHTML.Button
    item_classes = ['accordion-button']
    def __init__(self, key, base_name=None, **kw):
        self.base_name = base_name
        super().__init__([key], id=self.base_name+'-heading', **kw)
    def create_item(self, i, **kw):
        return super().create_item(i, type='button', data_bs_toggle='collapse', data_bs_target='#'+self.base_name+'-collapse')
class AccordionBody(Container):
    wrapper_classes = ['accordion-collapse', 'collapse']
    item = JHTML.Div
    item_classes = ['accordion-body']
    def __init__(self, key, parent_name=None, base_name=None, **kw):
        self.base_name = base_name
        super().__init__([key], id=self.base_name+'-collapse', data_bs_parent='#'+parent_name, **kw)
class Accordion(MenuComponent):
    wrapper_classes = ['accordion']
    item = JHTML.Div
    item_classes = ['accordion-item']
    header_classes = ['h2']
    def __init__(self, items, base_name=None, header_classes=None, **attrs):
        if base_name is None:
            base_name = 'accordion-' + short_uuid()
        self.base_name = base_name
        if isinstance(items, dict):
            items = items.items()
        self._active = None
        if header_classes is not None:
            self.header_classes = JHTML.manage_class(header_classes)
        super().__init__(items, id=self.base_name, **attrs)
    def create_item(self, item, cls=None, **kw):
        key, item = item
        item_id = self.base_name + "-" + short_uuid(3)
        cls = JHTML.manage_class(cls)
        if self._active is None:
            cls = cls + ['show']
            self._active = item_id
            header_cls = None
        else:
            header_cls = ['collapsed']

        header = AccordionHeader(key, base_name=item_id, item_attrs={'cls':header_cls}, wrapper_classes=self.header_classes)
        body = AccordionBody(item, parent_name=self.base_name, base_name=item_id, cls=cls, **kw)
        return super().create_item([header, body])

class OpenerHeader(Container):
    wrappers = {
        'wrapper':JHTML.Div,
        'item':Button
    }
    theme = {
        'wrapper':{'cls':['collapse-header']},
        'item':{'cls':['accordion-button']},
    }
    def __init__(self, key, base_name=None, **kw):
        self.base_name = base_name
        super().__init__([key], id=self.base_name+'-heading', **kw)
    def create_item(self, i, **kw):
        return super().create_item(i, type='button', data_bs_toggle='collapse', data_bs_target='#'+self.base_name+'-collapse')
class OpenerBody(Container):
    wrappers = {
        'wrapper': JHTML.Div,
        'item': JHTML.Div
    }
    theme = {
        'wrapper': {'cls': ['collapse']},
        'item': {'cls': ['collapse-body']},
    }
    def __init__(self, key, base_name=None, **kw):
        self.base_name = base_name
        super().__init__([key], id=self.base_name+'-collapse', **kw)#, data_bs_parent='#'+parent_name, **kw)
class Opener(MenuComponent):
    wrappers = {
        'wrapper':JHTML.Div,
        'item':JHTML.Div
    }
    theme = {
        'wrapper':{'cls':['opener']},
        'item':{'cls':['opener-item']},
        'header':{},
        'body':{}
    }
    def __init__(self, items, base_name=None, open=False, **attrs):
        if base_name is None:
            base_name = 'opener-' + short_uuid()
        self.base_name = base_name
        self.default_open=open
        if isinstance(items, dict):
            items = items.items()
        super().__init__(items, id=self.base_name, **attrs)
    def create_item(self, item, open=None, **kw):
        key, item = item
        item_id = self.base_name + "-" + short_uuid(3)
        if open is None:
            open = self.default_open

        ht = self.theme.get('header', {}).copy()
        ht['wrapper'] = self.merge_themes(
            ht.get('wrapper', {}),
            {'cls':['collapsed'] if not open else []}
        )
        ht['item'] = self.merge_themes(
            ht.get('item', {}),
            {'cls': ['collapsed'] if not open else []}
        )

        bt = self.theme.get('body', {}).copy()
        bt['wrapper'] = self.merge_themes(
            bt.get('wrapper', {}),
            {'cls': [] if not open else ['show']}
        )

        header = OpenerHeader(key, base_name=item_id, theme=ht)
        body = OpenerBody(item, base_name=item_id, theme=bt, **kw)
        return super().create_item([header, body])
class CardOpener(Opener):
    theme = {
        'wrapper': {'cls': ['opener', 'card', 'border-top-0']},
        'item': {'cls': ['opener-item']},
        "header": {
            'wrapper': {'cls': ['card-header', 'p-0', 'border-bottom-0', 'border-top']},
            'item': {'cls': ['text-dark', 'bg-transparent']}
        },
        "body": {'wrapper': {'cls': ['card-body']}}
    }

class Breadcrumb(MenuComponent):
    wrappers = {
        'wrapper':JHTML.Nav,
        'list':JHTML.Ol,
        'item':JHTML.Li
    }
    theme = {
        'wrapper': {'cls':[]},
        'list': {'cls':['breadcrumb']},
        'item': {'cls':['breadcrumb-item']}
    }

class CardBody(WrapperComponent):
    wrappers = {'wrapper':JHTML.Bootstrap.CardBody}
class CardHeader(WrapperComponent):
    wrappers = {'wrapper':JHTML.Bootstrap.CardHeader}
class CardFooter(WrapperComponent):
    wrappers = {'wrapper':JHTML.Bootstrap.CardFooter}
class Card(ComponentContainer):
    wrappers = {'wrapper':JHTML.Bootstrap.Card}
    components = {
        'header':CardHeader,
        'body':CardBody,
        'footer':CardFooter,
    }
    def __init__(self,
                 *args,
                 header=None,
                 body=None,
                 footer=None,
                 **attrs
                 ):
        if len(args) == 3:
            header, body, footer = args
        elif len(args) == 2:
            header, body = args
        elif len(args) == 1:
            body, = args
        elif len(args) > 0:
            raise NotImplementedError("too many body args")
        super().__init__(
            {
                'header':(header,),
                'body':(body,),
                'footer':(footer,)
            },
            **attrs
        )

class Modal(Container):
    wrapper_classes = ['modal', 'fade']
    subwrappers = [JHTML.Div, JHTML.Div]
    subwrapper_classes = [['modal-dialog', 'modal-dialog-centered'], ['modal-content']]
    def __init__(self,
                 header=None,
                 body=None,
                 footer=None,
                 id=None,
                 tabindex=-1,
                 **attrs
                 ):

        raise NotImplementedError("needs ComponentContainer update")
        items = []
        if header is not None:
            header = ModalHeader(header)
            items.append(header)
        self.header = header
        if body is not None:
            body = ModalBody(body)
            items.append(body)
        self.body = body
        if footer is not None:
            footer = ModalFooter(footer)
            items.append(footer)
        self.footer = footer
        if id is None:
            id = 'modal-'+short_uuid(6)
        self._id = id
        super().__init__(items, id=id, tabindex=tabindex, **attrs)

    trigger_class = JHTML.Bootstrap.Button
    def get_trigger(self, *items, trigger_class=None, data_bs_toggle='modal', data_bs_target=None, **attrs):
        if trigger_class is None:
            trigger_class = self.trigger_class
        if data_bs_target is None:
            data_bs_target = "#"+self._id
        return trigger_class(items, data_bs_toggle=data_bs_toggle, data_bs_target=data_bs_target, **attrs)
    @classmethod
    def close_button(self):
        return JHTML.Button(cls='btn-close', data_bs_dismiss='modal')
class ModalHeader(WrapperComponent):
    wrappers = {'wrapper':JHTML.Div}
    theme = {'wrapper':{'cls':['modal-header']}}
    def __init__(self, items, **attrs):
        items, attrs = self.manage_items(items, attrs)
        items.append(Modal.close_button())
        super().__init__(items, **attrs)
class ModalFooter(WrapperComponent):
    wrappers = {'wrapper':JHTML.Div}
    theme = {'wrapper':{'cls':['modal-footer']}}
class ModalBody(WrapperComponent):
    wrappers = {'wrapper':JHTML.Div}
    theme = {'wrapper':{'cls':['modal-body']}}

class Offcanvas(Container):
    wrapper_classes = ['offcanvas', 'ps-4', 'pt-5', 'pb-5']
    def __init__(self,
                 header=None,
                 body=None,
                 id=None,
                 tabindex=-1,
                 cls=None,
                 placement='start',
                 **attrs
                 ):
        raise NotImplementedError("needs ComponentContainer update")
        items = []
        if header is not None:
            header = OffcanvasHeader(header)
            items.append(header)
        self.header = header
        if body is not None:
            body = OffcanvasBody(body)
            items.append(body)
        self.body = body
        if id is None:
            id = 'offcanvas-'+short_uuid(6)
        self._id = id
        cls = JHTML.manage_class(cls) + ['offcanvas-' + placement]
        super().__init__(items, id=id, tabindex=tabindex, cls=cls, **attrs)

    trigger_class = JHTML.Bootstrap.Button
    def get_trigger(self, *items, trigger_class=None, data_bs_toggle='offcanvas', data_bs_target=None, **attrs):
        if trigger_class is None:
            trigger_class = self.trigger_class
        if data_bs_target is None:
            data_bs_target = "#"+self._id
        return trigger_class(items, data_bs_toggle=data_bs_toggle, data_bs_target=data_bs_target, **attrs)
    @classmethod
    def close_button(self):
        return JHTML.Button(cls='btn-close', data_bs_dismiss='offcanvas')
class OffcanvasHeader(WrapperComponent):
    wrappers = {'wrapper':JHTML.Div}
    theme = {'wrapper':{'cls':['offcanvas-header', 'm-2', 'border-bottom']}}
    def __init__(self, items, **attrs):
        items, attrs = self.manage_items(items, attrs)
        items.append(Offcanvas.close_button())
        super().__init__(items, **attrs)
class OffcanvasBody(WrapperComponent):
    wrappers = {'wrapper':JHTML.Div}
    theme = {'wrapper':{'cls':['offcanvas-body']}}

class Spacer(WrapperComponent):
    wrappers = {'wrapper':JHTML.Span}
    theme = {'wrapper':{'cls':['me-auto']}}
    def __init__(self, items=None, **kwargs):
        if items is None:
            items = []
        super().__init__(items, **kwargs)

ToastAPI = JHTML.JavascriptAPI.loader(
    init="""widget.el.toast = new context.bootstrap.Toast(widget.el)""",
    showToast="""
let id = widget.getAttribute("data-bs-target");
let el = document.querySelector(id);
el.toast.show();
"""
)
class ToastBody(WrapperComponent):
    wrappers = {
        'wrapper':JHTML.Div
    }
    theme = {
        'wrapper': {'cls':['toast-body']}
    }
    def __init__(self, items, include_controls=False, cls=None, **attrs):
        if include_controls:
            cls = JHTML.manage_class(cls) + ['d-flex']
            items, attrs = self.manage_items(items, attrs)
            items.extend([Spacer(), Toast.close_button()])
        super().__init__(items, cls=cls, **attrs)
class ToastHeader(WrapperComponent):
    wrappers = {
        'wrapper':JHTML.Div
    }
    theme = {
        'wrapper': {'cls':['toast-header']}
    }
    def __init__(self, items, include_controls=True, **attrs):
        if include_controls:
            items, attrs = self.manage_items(items, attrs)
            items.extend([Spacer(), Toast.close_button()])
        super().__init__(items, **attrs)
class Toast(WrapperComponent):
    wrappers = {
        'wrapper':JHTML.Div
    }
    theme = {
        'wrapper': {'cls':['toast']}
    }
    def __init__(self,
                 header=None,
                 body=None,
                 role='alert',
                 hidden=True,
                 cls=None,
                 id=None,
                 javascript_handles=None,
                 onevents=None,
                 **attrs
                 ):
        raise NotImplementedError("needs update")
        attrs['role'] = role
        items = []
        only_body = header is not None and body is None
        if only_body:
            body = header
            header = None
        if header is not None:
            items.append(ToastHeader(header))
        if body is not None:
            items.append(ToastBody(body, include_controls=only_body))
        if id is None:
            id = 'toast-'+short_uuid(6)
        self._id = id
        if not hidden:
            cls = JHTML.manage_class(cls) + ['show']
        if javascript_handles is None:
            javascript_handles = ToastAPI.load()
        if onevents is None:
            onevents = {'initialize':'init'}
        super().__init__(items, cls=cls, javascript_handles=javascript_handles, onevents=onevents, id=id, **attrs)

    trigger_class = JHTML.Bootstrap.Button
    def get_trigger(self, *items, trigger_class=None, data_bs_toggle='toast', data_bs_target=None, **attrs):
        if trigger_class is None:
            trigger_class = self.trigger_class
        if data_bs_target is None:
            data_bs_target = "#" + self._id
        return trigger_class(items,
                             data_bs_toggle=data_bs_toggle,
                             data_bs_target=data_bs_target,
                             javascript_handles=self.to_widget().javascript_handles,
                             event_handlers={"click":"showToast"},
                             **attrs
                             )
    @classmethod
    def close_button(self):
        return JHTML.Button(cls='btn-close', data_bs_dismiss='toast')
    def show(self):
        self.add_class('show')
        self.remove_class('hide')
    def hide(self):
        self.remove_class('show')
        self.add_class('hide')
class ToastContainer(WrapperComponent):
    wrappers = {
        'wrapper':JHTML.Div
    }
    theme = {
        'wrapper': {'cls':['toast-container']}
    }
    def __init__(self, items=None, **kwargs):
        if items is None:
            items = []
        super().__init__(items, **kwargs)
    def create_toast(self, header=None, body=None, hidden=False, **kwargs):
        toast = Toast(header=header, body=body, hidden=hidden, **kwargs)
        self.append(toast)
        return toast


#endregion

#region Misc

class Tooltip(ModifierComponent):
    modifiers = dict(
        data_bs_toggle="tooltip",
        data_bs_placement='top',
        javascript_handles={
        "update": """
if (widget.tooltip !== null && widget.tooltip !== undefined) {
    let title = widget.getAttribute('title');
    if (!(title && typeof title === 'object') || title !== this.tooltipTitle) {
        this.callHandler("init", event);
    } else {
        this.tooltip.update();
        this.el.setAttribute("title", "");
    }
}
                """,
        "init": """
let shown=false;
if (widget.tooltip !== null && widget.tooltip !== undefined) {
    if (widget.tooltip.tip !== null) {
        widget.tooltipTitle = null;
        let tip = widget.tooltip.tip;
        shown = tip && tip.classList.contains("show");
        widget.tooltip.dispose();
    }
}
widget.tooltip = new context.bootstrap.Tooltip(widget.el);
widget.tooltip._setContent = widget.tooltip.setContent;
function wrapperSetContent(tip) {
    widget.tooltip._setContent(tip);
    let title = widget.getAttribute('title');
    if (title && typeof title === 'object') {
        if (!(widget.tooltip.tip && title == widget.tooltipTitle)) {
            widget.tooltipTitle = title;
            let bwrap = tip.querySelector('.tooltip-inner');
            if (bwrap) {
                widget.create_child_view(title).then(
                  (view)=>{
                        while (bwrap.lastChild) {
                            bwrap.removeChild(bwrap.lastChild);
                        };
                        bwrap.appendChild(view.el)
                    }
                );
            }
        }
    }
    tip.classList.add('jhtml');
    return tip;
}
widget.tooltip.setContent = wrapperSetContent;
if (shown) {
    let trigger = widget.el.getAttribute("data-bs-trigger");
    if (trigger && trigger.split(' ').includes('click')) {
        widget.el.click();
    } else {
        widget.tooltip.show()
    }
};
""",
        "destroy": """
    if (widget.tooltip !== null && widget.tooltip !== undefined) {
        if (widget.tooltip.tip !== null) {
            widget.tooltip.dispose();
        }
    }"""
        },
        onevents={
            'initialize': "init",
            'view-change:data-bs-trigger': 'init',
            'view-change:data-bs-placement': 'init',
            'view-change:title': 'update',
            'remove': "destroy"
        }
    )
    def __init__(self, base=None, title="tooltip", data_bs_html=None, **kwargs):
        if not isinstance(title, str):
            if hasattr(title, 'tostring'):
                title = title.tostring()
                if data_bs_html is None:
                    data_bs_html = True
            elif hasattr(title, 'to_widget'):
                title = title.to_widget()
                if hasattr(title, 'elem'):
                    title = title.elem
                kwargs['onevents'] = dict(self.modifiers['onevents'], change='init')
                if data_bs_html is None:
                    data_bs_html = True
            elif isinstance(title, JupyterAPIs.get_widgets_api().Widget):
                if data_bs_html is None:
                    data_bs_html = True
            elif hasattr(title, 'to_tree'):
                title = title.to_tree().tostring()
                if data_bs_html is None:
                    data_bs_html = True
            else:
                raise TypeError('tooltip title must be a string')
        if data_bs_html is not None:
            kwargs['data_bs_html'] = data_bs_html
        super().__init__(base, title=title, **kwargs)
class Popover(ModifierComponent):
    modifiers = dict(
        data_bs_toggle="popover",
        data_bs_placement='top',
        data_bs_container='body',
        tabindex="0",
        javascript_handles={
            "update":"""
if (widget.popover !== null && widget.popover !== undefined) {
    let body = widget.getAttribute('data-bs-content');
    if (!(body && typeof body === 'object') || body !== this.popoverBody) {
        this.callHandler("init", event);
    } else {
        let title = widget.getAttribute('title');
        if (!(title && typeof title === 'object') || title !== this.popoverTitle) {
            this.callHandler("init", event);
        } else {
            this.popover.update();
            this.el.setAttribute("title", "")
        }
    }
}
            """,
            "init":"""
let shown=false;
if (widget.popover !== null && widget.popover !== undefined) {
    if (widget.popover.tip !== null) {
        widget.popoverBody = null;
        widget.popoverTitle = null;
        let tip = widget.popover.tip;
        shown = tip && tip.classList.contains("show");
        widget.popover.dispose();
    }
}
widget.popover = new context.bootstrap.Popover(widget.el);
widget.popover._oldGetTipElement = widget.popover.getTipElement;
function wrapperGetElement() {
    let tip = widget.popover._oldGetTipElement();
    let body = widget.getAttribute('data-bs-content');
    if (body && typeof body === 'object') {
        if (body !== widget.popoverBody) {
            widget.popoverBody = body;
            let bwrap = widget.popover.tip.querySelector('.popover-body');
            if (bwrap) {
                widget.create_child_view(body).then(
                  (view)=>{
                        while (bwrap.lastChild) {
                            bwrap.removeChild(bwrap.lastChild);
                        };
                        bwrap.appendChild(view.el)
                    }
                );
            }
        }
    }
    let title = widget.getAttribute('title');
    if (title && typeof title === 'object') {
        if (title !== widget.popoverTitle) {
            widget.popoverTitle = title;
            let bwrap = widget.popover.tip.querySelector('.popover-header');
            if (bwrap) {
                widget.create_child_view(title).then(
                  (view)=>{
                        while (bwrap.lastChild) {
                            bwrap.removeChild(bwrap.lastChild);
                        };
                        bwrap.appendChild(view.el)
                    }
                );
            }
        }
    }
    tip.classList.add('jhtml');
    return tip;
}
widget.popover.getTipElement = wrapperGetElement;
if (shown) {
    const triggers = widget.el.getAttribute("data-bs-trigger").split(' ');
    if (triggers.includes('click')) {
        widget.el.click();
    } else {
        widget.popover.show()
    }
};
""",
                            "destroy":"""
if (widget.popover !== null && widget.popover !== undefined) {
    if (widget.popover.tip !== null) {
        widget.popover.dispose();
    }
}"""},
        onevents={
            'initialize':"init",
            'view-change:data-bs-trigger':'init',
            'view-change:data-bs-placement':'init',
            'view-change:title': 'update',
            'view-change:data-bs-content':'update',
            'remove':"destroy"
        }
    )
    def __init__(self, base=None, body="", data_bs_trigger="hover focus", data_bs_html=None, title=None, **kwargs):
        if title is not None:
            if not isinstance(title, str):
                if hasattr(title, 'tostring'):
                    title = title.tostring()
                    if data_bs_html is None:
                        data_bs_html = True
                elif hasattr(title, 'to_tree'):
                    title = title.to_tree().tostring()
                    if data_bs_html is None:
                        data_bs_html = True
                elif isinstance(title, JupyterAPIs.get_widgets_api().Widget):
                    if data_bs_html is None:
                        data_bs_html = True
                elif hasattr(title, 'to_widget'):
                    title = title.to_widget()
                    if hasattr(title, 'elem'):
                        title = title.elem
                    if data_bs_html is None:
                        data_bs_html = True
                else:
                    raise TypeError('popover title must be a string')
            kwargs['title'] = title
        if not isinstance(body, str):
            if hasattr(body, 'tostring'):
                body = body.tostring()
                if data_bs_html is None:
                    data_bs_html = True
            elif hasattr(body, 'to_tree'):
                body = body.to_tree().tostring()
                if data_bs_html is None:
                    data_bs_html = True
            elif isinstance(body, JupyterAPIs.get_widgets_api().Widget):
                if data_bs_html is None:
                    data_bs_html = True
            elif hasattr(body, 'to_widget'):
                body = body.to_widget()
                if hasattr(body, 'elem'):
                    body = body.elem
                if data_bs_html is None:
                    data_bs_html = True
            else:
                raise TypeError('popover body must be a string')
        if data_bs_trigger is not None:
            kwargs['data_bs_trigger'] = data_bs_trigger
        if data_bs_html is not None:
            kwargs['data_bs_html'] = data_bs_html
        super().__init__(base, data_bs_content=body, **kwargs)

#endregion

#region Layouts

class LayoutItem(Component):
    wrapper = JHTML.Div
    properties = []
    def __init__(self, item, **attrs):
        super().__init__()
        self.item = item
        self.attrs = attrs
    @abc.abstractmethod
    def get_layout_styles(self, **kwargs):
        raise NotImplementedError("LayoutItem is an abstract class")
    def to_jhtml(self):
        attrs = self.attrs
        style = self.get_layout_styles()
        if 'style' in self.attrs:
            style = dict(attrs['style'], **style)
            attrs = attrs.copy()
            del attrs['style']
        wat = self.wrapper(
            self.item,
            style=style,
            **attrs
        )
        return wat
class Layout(Component):
    wrapper = JHTML.Div
    Item = LayoutItem
    def __init__(self, elements, wrapper=None, item_attrs=None, style=None, **attrs):
        super().__init__()
        if item_attrs is None:
            item_attrs = {}
        self.layout_settings, self.elements = self.setup_layout(elements, item_attrs)
        self._style = {} if style is None else style
        self.attrs = attrs
        if wrapper is None:
            wrapper = self.wrapper
        self.wrapper = wrapper
    def wrap_item(self, e, attrs):
        return self.Item(e, **attrs)
    def setup_layout(self, elements, item_attrs):
        return None, [self.wrap_item(e, item_attrs) for e in elements]
    @abc.abstractmethod
    def get_layout_styles(self, **kwargs):
        raise NotImplementedError("Layout is an abstract class")
    @property
    def styles(self):
        return dict(self._style, **self.get_layout_styles())
    def to_jhtml(self):
        return self.wrapper(
            *self.elements,
            style=self.styles,
            **self.attrs
        )

class GridItem(LayoutItem):
    def __init__(self, item,
                 row=None, col=None,
                 row_span=None, col_span=None,
                 alignment=None, justification=None,
                 **attrs
                 ):
        super().__init__(item, **attrs)
        self.row = row
        self.col = col
        self.row_span = row_span
        self.col_span = col_span
        self.alignment = alignment
        self.justification = justification
        self.attrs = attrs
    @classmethod
    def get_grid_styles(cls,
                           row=None, row_span=None,
                           col=None, col_span=None,
                           alignment=None, justification=None
                           ):
        settings = {}
        if row is not None:
            settings['grid-row-start'] = row
            if row_span is not None:
                settings['grid-row-end'] = 'span ' + str(row_span)
        if col is not None:
            settings['grid-column-start'] = col
            if col_span is not None:
                settings['grid-column-end'] = 'span ' + str(col_span)
        if alignment is not None:
            settings['align-self'] = alignment
        if justification is not None:
            settings['justify-self'] = justification
        return settings
    def get_layout_styles(self):
        return self.get_grid_styles(
            row=self.row, row_span=self.row_span,
            col=self.col, col_span=self.col_span,
            alignment=self.alignment, justification=self.justification,
        )
class Grid(Layout):
    Item = GridItem
    def __init__(self, elements,
                 rows=None, cols=None,
                 alignment=None, justification=None,
                 row_spacing=None, col_spacing=None,
                 item_attrs=None,
                 row_height='auto',
                 column_width='1fr',
                 **attrs
                 ):
        super().__init__(elements, item_attrs=item_attrs, **attrs)
        if rows is None:
            rows = self.layout_settings['rows']
        if cols is None:
            cols = self.layout_settings['cols']
        self.rows = rows
        self.cols = cols
        self.alignment = alignment
        self.justification = justification
        self.row_gaps = row_spacing
        self.col_gaps = col_spacing
        self.row_height = row_height
        self.col_width = column_width
    def setup_layout(self, grid, attrs):
        elements = []
        nrows = 0
        ncols = 0
        for i, row in enumerate(grid):
            for j, el in enumerate(row):
                if el is None:
                    continue
                elem = self.wrap_item(el, dict(attrs, row=i+1, col=j+1))
                n = elem.row
                if elem.row_span is not None:
                    n += elem.row_span
                if n > nrows:
                    nrows = n
                m = elem.col
                if elem.col_span is not None:
                    m += elem.col_span
                if m > ncols:
                    ncols = m
                elements.append(elem)
        return {'rows':nrows, 'cols':ncols}, elements
    def wrap_item(self, e, attrs):
        if not isinstance(e, self.Item):
            e = self.Item(e, **attrs)
        elif hasattr(e, 'items'):
            body = e['body']
            e = dict(e)
            del e['body']
            e = self.Item(body, **dict(attrs, **e))
        else:
            if e.row is None:
                e.row = attrs['row']
            if e.col is None:
                e.col = attrs['col']
        return e
    @classmethod
    def get_grid_styles(cls,
                        rows=None,  cols=None,
                        alignment=None, justification=None,
                        row_gap=None, col_gap=None,
                        row_height='1fr', col_width='1fr'
                        ):
        settings = {'display':'grid'}
        if rows is not None:
            if isinstance(row_height, str) and ' ' not in row_height:
                settings['grid-template-rows'] = 'repeat({rows}, {height})'.format(rows=rows, height=row_height) if isinstance(rows, int) else rows
            elif isinstance(row_height, str):
                settings['grid-template-rows'] = row_height
            else:
                settings['grid-template-rows'] = " ".join(row_height)

        if cols is not None:
            if isinstance(col_width, str) and ' ' not in col_width:
                settings['grid-template-columns'] = 'repeat({cols}, {width})'.format(cols=cols, width=col_width) if isinstance(cols, int) else cols
            elif isinstance(col_width, str):
                settings['grid-template-columns'] = col_width
            else:
                settings['grid-template-columns'] = " ".join(col_width)
        if alignment is not None:
            settings['align-items'] = alignment
        if justification is not None:
            settings['justify-items'] = justification
        if row_gap is not None:
            settings['row-gap'] = row_gap
        if col_gap is not None:
            settings['column-gap'] = col_gap
        return settings
    def get_layout_styles(self):
        return self.get_grid_styles(
            rows=self.rows, cols=self.cols,
            alignment=self.alignment, justification=self.justification,
            row_gap=self.row_gaps, col_gap=self.col_gaps,
            row_height=self.row_height, col_width=self.col_width
        )

class TableItem(GridItem):
    def __init__(self, item,
                 row=None, col=None,
                 row_span=None, col_span=None,
                 alignment=None, justification=None,
                 header=False,
                 **attrs
                 ):
        super().__init__(
            item,
            row=row, col=col,
            row_span=row_span, col_span=col_span,
            alignment=alignment, justification=justification,
            **attrs
        )
        self.header = header
    def wrapper(self, item, **kwargs):
        if self.header:
            return JHTML.TableHeading(item, **kwargs)
        else:
            return JHTML.TableItem(item,  **kwargs)
class Table(Grid):
    Item = TableItem
    def __init__(
            self,
            elements,
            rows=None, cols=None,
            alignment=None, justification=None,
            row_spacing=None, col_spacing=None,
            item_attrs=None,
            row_height='1fr',
            column_width='1fr',
            table_headings=None,
            striped=True,
            **attrs
    ):
        self.headings = table_headings
        self.striped = striped
        super().__init__(
            elements,
            rows=rows, cols=cols,
            alignment=alignment, justification=justification,
            row_spacing=row_spacing, col_spacing=col_spacing,
            item_attrs=item_attrs,
            row_height=row_height,
            column_width=column_width,
            **attrs
        )
    def wrapper(self, *elems, cls=None, **attrs):
        if self.striped:
            cls = ['table', 'table-striped'] + JHTML.manage_class(cls)
        else:
            cls = ['table'] + JHTML.manage_class(cls)
        if self.headings is not None:
            elems = [
                JHTML.TableHeader(elems[0], display='contents'),
                JHTML.TableBody(*elems[1:], display='contents')
            ]
        else:
            elems = [JHTML.TableBody(*elems, display='contents')]
        return JHTML.Table(
            *elems,
            cls=cls,
            **attrs
        )

    def setup_layout(self, grid, attrs):
        rows = []
        nrows = 0
        ncols = 0
        has_header = self.headings is not None
        if has_header:
            header = JHTML.TableRow(*[
                self.wrap_item(el, dict(attrs, row=0, col=j+1, header=True))
                for j,el in enumerate(self.headings)
                ],
                display='contents'
            )
            rows.append(header)
            nrows = 1
        for i, row in enumerate(grid):
            tr = []
            if has_header:
                i += 1
            for j, el in enumerate(row):
                if el is None:
                    continue
                elem = self.wrap_item(el, dict(attrs, row=i+1, col=j+1))
                n = elem.row
                if elem.row_span is not None:
                    n += elem.row_span
                if n > nrows:
                    nrows = n
                m = elem.col
                if elem.col_span is not None:
                    m += elem.col_span
                if m > ncols:
                    ncols = m
                tr.append(elem)
            rows.append(JHTML.TableRow(*tr, display='contents'))
        return {'rows':nrows, 'cols':ncols}, rows


class FlexItem(LayoutItem):
    def __init__(self,
                 item,
                 order=None, grow=None,
                 shrink=None, basis=None,
                 alignment=None,
                 **attrs
                 ):
        super().__init__(item, **attrs)
        self.order = order
        self.grow = grow
        self.shrink = shrink
        self.basis = basis
        self.alignment = alignment
    @classmethod
    def get_flex_styles(cls,
                        order=None, grow=None,
                        shrink=None, basis=None,
                        alignment=None
                        ):
        settings = {}
        if order is not None:
            settings['flex-order'] = order
        if grow is not None:
            settings['flex-grow'] = grow
        if shrink is not None:
            settings['flex-shrink'] = shrink
        if basis is not None:
            settings['flex-basis'] = basis
        if alignment is not None:
            settings['align-self'] = alignment
        return settings
    def get_layout_styles(self):
        return self.get_flex_styles(
            order=self.order, grow=self.grow,
            shrink=self.shrink, basis=self.basis,
            alignment=self.alignment
        )
class Flex(Layout):
    Item = FlexItem
    def __init__(self,
                 elements,
                 direction=None, wrap=None,
                 alignment=None, justification=None,
                 content_alignment=None,
                 **attrs
                 ):
        super().__init__(elements, **attrs)
        self.direction = direction
        self.wrap = wrap
        self.content_alignment = content_alignment
        self.alignment = alignment
        self.justification = justification
    @classmethod
    def get_flex_styles(cls,
                        direction=None, wrap=None,
                        alignment=None, justification=None,
                        content_alignment=None
                        ):
        settings = {'display': 'flex'}
        if direction is not None:
            settings['flex-direction'] = direction
        if wrap is not None:
            settings['flex-wrap'] = wrap
        if alignment is not None:
            settings['align-items'] = alignment
        if justification is not None:
            settings['justify-content'] = justification
        if content_alignment is not None:
            settings['align-content'] = content_alignment
        return settings
    def get_layout_styles(self):
        return self.get_flex_styles(
            direction=self.direction, wrap=self.wrap,
            alignment=self.alignment, justification=self.justification,
            content_alignment=self.content_alignment
        )

#endregion


class GenericDisplay(WidgetInterface):
    def __init__(self, obj):
        self.obj = obj
    def to_widget(self):
        if hasattr(self.obj, 'to_widget'):
            res = self.obj.to_widget()
        else:
            res = JHTML.OutputArea(autoclear=True)
            res.show_buffered(self.obj)
            # with res:
            # res.show_buffered(self.obj)
            # with res:
            #     if hasattr(self.obj, '_ipython_display_'):
            #         self.obj._ipython_display_()
            #     else:
            #         JupyterAPIs.get_display_api().display(self.obj)
        return res

class ResultTypes:
    NoResult = "NoResult"
class DelayedResult(WidgetInterface):

    NoResult = ResultTypes.NoResult

    def __init__(self, func, *args,
                 output=None,
                 callback=None,
                 parent=None,
                 **kwargs
                 ):
        self.output = self.get_output_area(output)
        self.caller = (func, args, kwargs)
        self._thread = None
        self.result = self.NoResult
        self.error = None
        self.callback = callback
        self.parent = parent

    def get_output_area(self, output=None):
        if output is None:
            output = JHTML.OutputArea()
        return output

    def __enter__(self):
        self.output.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.output.__exit__()

    def _run(self):
        # with self.output:
        try:
            res = self.caller[0](*self.caller[1], **self.caller[2], runner=self)
        except Exception as e:
            self.error = e
            with self.output:
                raise
        else:
            self.result = res
            # with self.output:
            try:
                if hasattr(res, 'to_widget'):
                    res = res.to_widget()
                self.output.show_buffered(res)
                # with self.output:
                #     if hasattr(res, '_ipython_display_'):
                #         res._ipython_display_()
                #     else:
                #         JupyterAPIs.get_display_api().display(res)
            except:
                with self.output:
                    raise

        if self.callback is not None:
            self.callback(self.result, self.error, self)

    def start_process(self):
        if self._thread is None:
            self._thread = threading.Thread(target=self._run)
            self._thread.start()
        return self._thread

    def to_widget(self):
        self.start_process()
        return self.output



from .GeometricTransformation import *
from .TransformationFunction import *
from .AffineTransform import *
from .TranslationTransform import *
from .RotationTransform import *

__all__ = []
from .GeometricTransformation import __all__ as exposed
__all__ += exposed
from .TransformationFunction import __all__ as exposed
__all__ += exposed
from .AffineTransform import __all__ as exposed
__all__ += exposed
from .TranslationTransform import __all__ as exposed
__all__ += exposed
from .RotationTransform import __all__ as exposed
__all__ += exposed


"""Implements calculation of Euler angles, Euler matrices, etc

"""

import numpy as np

__all__ = [
    "euler_angles",
    "euler_matrix"
]

# I pulled a bunch of Euler matrices for diferent conventions out of Mathematica with the
# code Resources/Mathematica/eulerSolver.wl
# Any new orientation can be easily added like that
# We could do some metaprogramming thing with eval(orientation+_mat)
# but I think it's better to just hard code the choices the dispatcher maps

#### ZYZ
def zyz_mat(c, s):
    return [
        [ c[0]*c[1]*c[2]+-s[0]*s[2], -c[2]*s[0]+-c[0]*c[1]*s[2], c[0]*s[1] ],
        [ c[1]*c[2]*s[0]+c[0]*s[2] , c[0]*c[2]+-c[1]*s[0]*s[2] , s[0]*s[1] ],
        [ -c[2]*s[1]               , s[1]*s[2]                 , c[1]      ]
    ]
def zyz_angles(basis):
    return [
        -np.arccos(-basis[2][0]*1/(np.sqrt(1+-(basis[2][2])**(2)))),
        np.arctan2(basis[2][2], -np.sqrt(1+-(basis[2][2])**(2))),
        -np.arccos(basis[0][2]*1/(np.sqrt(1+-(basis[2][2])**(2))))
    ]

##### XYZ
def xyz_mat(c, s):
    return [
        [ c[1]*c[2]                , -c[1]*s[2]               , s[1]       ],
        [ c[2]*s[0]*s[1]+c[0]*s[2] , c[0]*c[2]+-s[0]*s[1]*s[2], -c[1]*s[0] ],
        [ -c[0]*c[2]*s[1]+s[0]*s[2], c[2]*s[0]+c[0]*s[1]*s[2] , c[0]*c[1]  ]
    ]
def xyz_angles(basis):
    return [
        -np.arccos(-1/(np.sqrt((basis[0][0])**(2)+(basis[0][1])**(2)))*basis[2][2]),
        -np.arccos(-np.sqrt((basis[0][0])**(2)+(basis[0][1])**(2))),
        np.arctan2(-basis[0][0], basis[0][1])
    ]

##### XZY
def xzy_mat(c, s):
    return  [
        [ c[1]*c[2]                , -s[1]    , c[1]*s[2]                 ],
        [ c[0]*c[2]*s[1]+s[0]*s[2] , c[0]*c[1], -c[2]*s[0]+c[0]*s[1]*s[2] ],
        [ c[2]*s[0]*s[1]+-c[0]*s[2], c[1]*s[0], c[0]*c[2]+s[0]*s[1]*s[2]  ]
    ]
def xzy_angles(basis):
    return [
        -np.arccos(-1/(np.sqrt(1+-(basis[0][1])**(2)))*basis[1][1]),
        np.arctan2(-np.sqrt(1+-(basis[0][1])**(2)), -basis[0][1]),
        -np.arccos(-basis[0][0]*1/(np.sqrt(1+-(basis[0][1])**(2))))
    ]

##### ZYX
def zyx_mat(c, s):
    return [
        [ c[0]*c[1], -c[2]*s[0]+c[0]*s[1]*s[2], c[0]*c[2]*s[1]+s[0]*s[2]  ],
        [ c[1]*s[0], c[0]*c[2]+s[0]*s[1]*s[2] , c[2]*s[0]*s[1]+-c[0]*s[2] ],
        [ -s[1]    , c[1]*s[2]                , c[1]*c[2]                 ]
    ]
def zyx_angles(basis):
    return [
        np.arctan2(-basis[0][0], -basis[1][0]),
        -np.arccos(-np.sqrt((basis[0][0])**(2)+(basis[1][0])**(2))),
        -np.arccos(-1/(np.sqrt((basis[0][0])**(2)+(basis[1][0])**(2)))*basis[2][2])
    ]

### Define a dispatch map for these functions

euler_mat_map = {
    "zyz" : zyz_mat,
    "xyz" : xyz_mat,
    "zyx" : zyx_mat
}

# After we let Mathematica do the heavy lifting of actually generating the matrices
# we then simply take their sines and cosines

def euler_matrix(angles, ordering="xyz"):
    """Returns the Euler matrix for the specified angles

    :param angles:
    :type angles:
    :param ordering: the order in which the rotations should be performed
    :type ordering:
    """
    ordering = ordering.lower()
    if ordering in euler_mat_map:
        mat_gen = euler_mat_map[ordering]
    else:
        raise KeyError("Euler matrix for ordering '{}' not yet supported".format(ordering))
    c = np.cos(angles)
    s = np.sin(angles)
    return mat_gen(c, s)

def euler_angles(basis, ordering="xyz"):
    """Calculates the Euler angles for the basis

    :param basis: the basis to get the Euler angles for
    :type basis: np.ndarray
    """
    if ordering in euler_mat_map:
        ang_gen = euler_mat_map[ordering]
    else:
        raise KeyError("Euler matrix for orientation '{}' not yet supported".format(ordering))
    return ang_gen(basis)


import collections

import numpy as np
import itertools
from .. import Numputils as nput
from .. import Iterators as itut
from ..Graphs import EdgeGraph

from .Internals import canonicalize_internal

__all__ = [
    "zmatrix_unit_convert",
    "zmatrix_indices",
    "num_zmatrix_coords",
    "zmatrix_embedding_coords",
    "set_zmatrix_embedding",
    "enumerate_zmatrices",
    "extract_zmatrix_internals",
    "parse_zmatrix_string",
    "format_zmatrix_string",
    "validate_zmatrix",
    "chain_zmatrix",
    "center_bound_zmatrix",
    "spoke_zmatrix",
    # "methyl_zmatrix",
    # "ethyl_zmatrix",
    "attached_zmatrix_fragment",
    "functionalized_zmatrix",
    "add_missing_zmatrix_bonds",
    "bond_graph_zmatrix",
    "reindex_zmatrix",
    "sort_complex_attachment_points",
    "complex_zmatrix"
]


def zmatrix_unit_convert(zmat, distance_conversion, angle_conversion=None, rad2deg=False, deg2rad=False):
    zm2 = np.asanyarray(zmat)
    if zm2 is zmat: zm2 = zm2.copy()

    zm2[..., :, 0] *= distance_conversion
    if angle_conversion is None:
        if deg2rad:
            zm2[..., :, 1] = np.deg2rad(zm2[..., :, 1])
            zm2[..., :, 2] = np.deg2rad(zm2[..., :, 2])
        elif rad2deg:
            zm2[..., :, 1] = np.rad2deg(zm2[..., :, 1])
            zm2[..., :, 2] = np.rad2deg(zm2[..., :, 2])
    else:
        zm2[..., :, 1] *= angle_conversion
        zm2[..., :, 2] *= angle_conversion

    return zm2

def zmatrix_indices(zmat, coords, strip_embedding=True):
    base_coords = [canonicalize_internal(c) for c in extract_zmatrix_internals(zmat, strip_embedding=strip_embedding)]
    return [
        base_coords.index(canonicalize_internal(c))
        for c in coords
    ]

emb_pos_map = [
    (0,1),
    (0,2),
    (0,3),
    None,
    (1,2),
    (1,3),
    None,
    None,
    (2,3)
]
def zmatrix_embedding_coords(zmat_or_num_atoms, array_inds=False):
    if array_inds:
        base_inds = zmatrix_embedding_coords(zmat_or_num_atoms, array_inds=False)
        return [emb_pos_map[i] for i in base_inds]
    else:
        if not nput.is_int(zmat_or_num_atoms):
            zmat_or_num_atoms = len(zmat_or_num_atoms) + (1 if len(zmat_or_num_atoms[0]) == 3 else 0)
        n: int = zmat_or_num_atoms

        if n < 1:
            return []
        elif n == 1:
            return [0, 1, 2]
        elif n == 2:
            return [0, 1, 2, 4, 5]
        else:
            return [0, 1, 2, 4, 5, 8]

def num_zmatrix_coords(zmat_or_num_atoms, strip_embedding=True):
    if not nput.is_int(zmat_or_num_atoms):
        zmat_or_num_atoms = len(zmat_or_num_atoms) + (1 if len(zmat_or_num_atoms[0]) == 3 else 0)
    n: int = zmat_or_num_atoms

    return (n*3) - (
        0
            if not strip_embedding else
        len(zmatrix_embedding_coords(n))
    )

def _zmatrix_iterate(coords, natoms=None,
                     include_origins=False,
                     canonicalize=True,
                     deduplicate=True,
                     allow_completions=False
                     ):
    # TODO: this fixes an atom ordering, to change that up we'd need to permute the initial coords...
    if canonicalize:
        coords = [tuple(reversed(canonicalize_internal(s))) for s in coords]

    if deduplicate:
        dupes = set()
        _ = []
        for c in coords:
            if c in dupes: continue
            _.append(c)
            dupes.add(c)
        coords = _

    if include_origins:
        if (1, 0) not in coords:
            coords = [(1, 0)] + coords
        if (2, 1) not in coords and (2, 0) not in coords:
            if (2, 1, 0) in coords:
                coords = [(2, 1)] + coords
            else:
                coords = [(2, 0)] + coords
        if (2, 0) in coords and (2, 0, 1) not in coords: # can this happen?
            coords.append((2,1,0))

    if natoms is None:
        all_atoms = {i for s in coords for i in s}
        natoms = len(all_atoms)

    dihedrals = [k for k in coords if len(k) == 4]
    all_dihedrals = [
        (i, j, k, l)
        for (i, j, k, l) in dihedrals
        if i > j and i > k and i > l
    ]

    # need to iterate over all N-2 choices of dihedrals (in principle)...
    # should first reduce over consistent sets
    if not allow_completions:
        dihedrals = [
            (i,j,k,l) for i,j,k,l in dihedrals
            if (i,j) in coords and (i,j,k) in coords
            # if (
            #         any(x in coords or tuple(reversed(x)) in coords for x in [(i,j), (l,k)])
            #         and any(x in coords or tuple(reversed(x)) in coords for x in [(i,j,k), (l,k,j)])
            # )
        ]

    embedding = [
        x for x in [(2, 0, 1), (2, 1, 0)]
        if x in coords
    ]

    # we also will want to sample from dihedrals that provide individual atoms
    atom_diheds = [[] for _ in range(natoms)]
    for n,(i,j,k,l) in enumerate(dihedrals):
        atom_diheds[i].append((i,j,k,l))

    # completions = []
    # if allow_completions:
    #     for d in all_dihedrals:
    #         if d in dihedrals: continue
    #         completions.extend([d[:2], d[:3]])
    #
    #     c_set = set()
    #     for d in dihedrals:
    #         c_set.add(d[:2])
    #         c_set.add(d[:3])
    #     coord_pairs = [
    #         (c[:2],c[:3])
    #         for
    #     ]
    #     for d in all_dihedrals:
    #         if d in dihedrals: continue
    #         completions.extend([d[:2], d[:3]])

    for dihed_choice in itertools.product(embedding, *atom_diheds[3:]):
        emb, dis = dihed_choice[0], dihed_choice[1:]
        yield (
            (0, -1, -1, -1),
            (1, 0, -1, -1),
            emb + (-1,)
        ) + dis

def enumerate_zmatrices(coords, natoms=None,
                        allow_permutation=True,
                        include_origins=False,
                        canonicalize=True,
                        deduplicate=True,
                        preorder_atoms=True,
                        allow_completions=False
                        ):
    if canonicalize:
        coords = [tuple(reversed(canonicalize_internal(s))) for s in coords]

    if deduplicate:
        dupes = set()
        _ = []
        for c in coords:
            if c in dupes: continue
            _.append(c)
            dupes.add(c)
        coords = _

    if natoms is None:
        all_atoms = {i for s in coords for i in s}
        natoms = len(all_atoms)

    if preorder_atoms:
        counts = itut.counts(itertools.chain(*coords))
        max_order = list(sorted(range(natoms), key=lambda k:-counts[k]))
    else:
        max_order = np.arange(natoms)

    for atoms in (
            itertools.permutations(max_order)
                if allow_permutation else
            [max_order]
    ):
        atom_perm = np.argsort(atoms)
        perm_coords = [
            tuple(reversed(canonicalize_internal([atom_perm[c] for c in crd])))
            for crd in coords
        ]
        for zm in _zmatrix_iterate(perm_coords,
                                   natoms=natoms,
                                   include_origins=include_origins,
                                   canonicalize=False,
                                   deduplicate=False,
                                   allow_completions=allow_completions
                                   ):
            yield [
                [atoms[c] if c >= 0 else c for c in z]
                for z in zm
            ]

def extract_zmatrix_internals(zmat, strip_embedding=True, canonicalize=True):
    specs = []
    if len(zmat[0]) == 3:
        return extract_zmatrix_internals(
            [[0, -1, -1, -1]]
            + [
                [i + 1] + list(z)
                for i, z in enumerate(zmat)
            ],
            strip_embedding=strip_embedding,
            canonicalize=canonicalize
        )
        # zmat = np.asanyarray(zmat)
        # return np.delete(zmat.flatten(), zmatrix_embedding_coords(len(zmat)))
    else:
        for n,row in enumerate(zmat):
            if strip_embedding and n == 0: continue
            if canonicalize:
                coord = canonicalize_internal(row[:2])
            else:
                coord = tuple(row[:2])
            specs.append(coord)
            if strip_embedding and n == 1: continue
            if canonicalize:
                coord = canonicalize_internal(row[:3])
            else:
                coord = tuple(row[:3])
            specs.append(coord)
            if strip_embedding and n == 2: continue
            if canonicalize:
                coord = canonicalize_internal(row[:4])
            else:
                coord = tuple(row[:4])
            specs.append(coord)
    return specs

scan_spec = collections.namedtuple('scan_spec', ['value', 'steps', 'amount'])
def _prep_var_spec(v):
    if len(v) == 1:
        return float(v[0])
    elif len(v) == 2 and v[1] in {'f', 'F'}:
        return scan_spec(float(v[0]), -1, 0)
    elif len(v) == 3:
        return scan_spec(float(v[0]), int(v[1]), float(v[2]))
    else:
        raise ValueError(f"can't parse var spec {v}")
def parse_zmatrix_string(zmat, units="Angstroms", in_radians=False,
                         keep_variables=False,
                         variables=None,
                         dialect='gaussian'):
    from ..Data import AtomData, UnitsData
    # we have to reparse the Gaussian Z-matrix...

    possible_atoms = {d["Symbol"][:2] for d in AtomData.data.values()}

    atoms = []
    ordering = []
    coords = []
    # vars = {}

    if "Variables:" in zmat:
        zmat, vars_block = zmat.split("Variables:", 1)
    else:
        zmat = zmat.split("\n\n", 1)
        if len(zmat) == 1:
            zmat = zmat[0]
            vars_block = ""
        else:
            zmat, vars_block = zmat
    bits = [b.strip() for b in zmat.split() if len(b.strip()) > 0]

    coord = []
    ord = []
    complete = False
    last_complete = -1
    last_idx = len(bits) - 1
    for i, b in enumerate(bits):
        d = (i - last_complete) - 1
        m = d % 2
        if d == 0:
            atoms.append(b)
        elif m == 1:
            b = int(b)
            if b > 0: b = b - 1
            ord.append(b)
        elif m == 0:
            coord.append(b)

        terminal = (
                i == last_idx
                or i in {0, 3, 8}
                or (i > 8 and (i - 9) % 7 == 6)
        )
        # atom_q = bits[i + 1][:2] in possible_atoms
        if terminal:
            last_complete = i
            ord = [len(ordering)] + ord + [-1] * (3 - len(ord))
            coord = coord + [0] * (3 - len(coord))
            ordering.append(ord)
            coords.append(coord)
            ord = []
            coord = []

    split_vars = [
        vb.strip().replace("=", " ").split()
        for vb in vars_block.split("\n")
    ]
    # split_pairs = [s for s in split_pairs if len(s) == 2]

    if dialect != "gaussian":
        raise NotImplementedError(f"unsupported z-matrix dialect '{dialect}'")
    vars = {
        v[0]:_prep_var_spec(v[1:])
        for v in split_vars
        if len(v) > 0
    }
    if variables is not None:
        vars.update(variables)

    # ordering = [
    #     [i] + o
    #     for i, o in enumerate(ordering)
    # ]

    if not keep_variables:
        vals = {
            k:(
                v.value
                    if not nput.is_numeric(v) else
                v
            )
            for k, v in vars.items()
        }
        coords = [
            [
                vals[x] if x in vals else float(x)
                for x in c
            ]
            for c in coords
        ]

        # convert book angles into sensible dihedrals...
        # actually...I think I don't need to do anything for this?
        ordering = np.array(ordering)[:, :4]

        coords = np.array(coords)
        coords[:, 0] *= UnitsData.convert(units, "BohrRadius")
        coords[:, 1] = coords[:, 1] if in_radians else np.deg2rad(coords[:, 1])
        coords[:, 2] = coords[:, 2] if in_radians else np.deg2rad(coords[:, 2])
        return (atoms, ordering, coords)
    else:
        return (atoms, ordering, coords), vars

def format_zmatrix_string(atoms, zmat, ordering=None, units="Angstroms",
                          in_radians=False,
                          float_fmt="{:11.8f}",
                          index_padding=1,
                          variables=None,
                          variable_modifications=None,
                          distance_variable_format="r{i}",
                          angle_variable_format="a{i}",
                          dihedral_variable_format="d{i}"
                          ):
    from ..Data import UnitsData
    if ordering is None:
        if len(zmat) == len(atoms):
            zmat = zmat[1:]
        ordering = [
            [z[0], z[2], z[4]]
            if i > 1 else
            [z[0], z[2], -1]
            if i > 0 else
            [z[0], -1, -1]
            for i, z in enumerate(zmat)
        ]
        zmat = [
            [z[1], z[3], z[5]]
            if i > 1 else
            [z[1], z[3], -1]
            if i > 0 else
            [z[1], -1, -1]
            for i, z in enumerate(zmat)
        ]

    if isinstance(zmat, np.ndarray):
        zmat = zmat.copy()
        zmat[:, 0] *= UnitsData.convert("BohrRadius", units)
        zmat[:, 1] = zmat[:, 1] if in_radians else np.rad2deg(zmat[:, 1])
        zmat[:, 2] = zmat[:, 2] if in_radians else np.rad2deg(zmat[:, 2])
        zmat = zmat.tolist()
    else:
        cr = UnitsData.convert("BohrRadius", units)
        zmat = [
            [
                r * cr if nput.is_numeric(r) else r,
                np.rad2deg(a) if not in_radians and nput.is_numeric(a) else a,
                np.rad2deg(d) if not in_radians and nput.is_numeric(d) else d
            ]
            for r, a, d in zmat
        ]

    if variables is True:
        variables = {}
        _ = []
        for i,(r,a,d) in enumerate(zmat):
            s = []
            vr = distance_variable_format.format(i=i)
            variables[vr] = r
            s.append(vr)
            if i > 0:
                va = angle_variable_format.format(i=i)
                variables[va] = a
                s.append(va)
            else:
                s.append("")
            if i > 1:
                vd = dihedral_variable_format.format(i=i)
                variables[vd] = d
                s.append(vd)
            else:
                s.append("")
            _.append(s)
        zmat = _

    includes_atom_list = len(ordering[0]) == 4
    if not includes_atom_list:
        if len(ordering) < len(atoms):
            ordering = [[-1, -1, -1, -1]] + list(ordering)
        if len(zmat) < len(atoms):
            zmat = [[-1, -1, -1]] + list(zmat)

    if variable_modifications is not None:
        if variables is None:
            variables = {}
        includes_atom_list = len(ordering[0]) == 4
        for i,(x,r,a,d) in enumerate(ordering):
            for k,fmt,j in [
                [(x, r), distance_variable_format, 0],
                [(r, x), distance_variable_format, 0],
                [(x, r, a), angle_variable_format, 1],
                [(a, r, x), angle_variable_format, 1],
                [(x, r, a, d), dihedral_variable_format, 2],
                [(d, a, r, x), dihedral_variable_format, 2],
            ]:
                if k in variable_modifications:
                    vr = fmt.format(i=i)
                    val = variables.get(vr, zmat[i][j])
                    if not isinstance(val, str):
                        val = float_fmt.format(val)
                    variables[vr] = val + " " + variable_modifications[k]
                    zmat[i][j] = vr
                    break

    zmat = [
        ["", "", ""]
        if i == 0 else
        [z[0], "", ""]
        if i == 1 else
        [z[0], z[1], ""]
        if i == 2 else
        [z[0], z[1], z[2]]
        for i, z in enumerate(zmat)
    ]
    zmat = [
        [
            float_fmt.format(x)
                if not isinstance(x, str) else
            x
            for x in zz
        ]
        for zz in zmat
    ]
    if includes_atom_list:
        ord_list = [o[0] for o in ordering]
        atom_order = np.argsort(ord_list)
        atoms = [atoms[o] for o in ord_list]
        ordering = [
            ["", "", ""]
              if i == 0 else
            [atom_order[z[1]], "", ""]
              if i == 1 else
            [atom_order[z[1]], atom_order[z[2]], ""]
              if i == 2 else
            [atom_order[z[1]], atom_order[z[2]], atom_order[z[3]]]
              for i, z in enumerate(ordering)
        ]
    else:
        ordering = [
            ["", "", ""]
            if i == 0 else
                [z[0], "", ""]
            if i == 1 else
                [z[0], z[1], ""]
            if i == 2 else
                [z[0], z[1], z[2]]
            for i, z in enumerate(ordering)
        ]
    ordering = [
        ["{:.0f}".format(x + index_padding) if not isinstance(x, str) else x for x in zz]
        for zz in ordering
    ]

    max_at_len = max(len(a) for a in atoms)

    nls = [
        max([len(xyz[i]) for xyz in ordering])
        for i in range(3)
    ]
    zls = [
        max([len(xyz[i]) for xyz in zmat])
        for i in range(3)
    ]

    fmt_string = f"{{a:<{max_at_len}}} {{n[0]:>{nls[0]}}} {{r[0]:>{zls[0]}}} {{n[1]:>{nls[1]}}} {{r[1]:>{zls[1]}}} {{n[2]:>{nls[2]}}} {{r[2]:>{zls[2]}}}"
    zm = "\n".join(
        fmt_string.format(
            a=a,
            n=n,
            r=r
        )
        for a, n, r in zip(atoms, ordering, zmat)
    )
    if variables is not None:
        max_symbol_len = max(len(s.split()[0]) for s in variables)
        variables = {
            k: v if isinstance(v, str) else float_fmt.format(v)
            for k, v in variables.items()
        }
        max_v_len = max(len(s) for s in variables.values())
        variables_fmt = f" {{:<{max_symbol_len}}} = {{:>{max_v_len}}}"
        variables_block = "\n".join(
            variables_fmt.format(k, v if isinstance(v, str) else float_fmt.format(v))
            for k,v in variables.items()
        )
        zm = zm + "\nVariables:\n" + variables_block

    return zm


def validate_zmatrix(ordering,
                     allow_reordering=True,
                     ensure_nonnegative=True,
                     # raise_exception=False
                     ):
    proxy_order = np.array([o[0] for o in ordering])
    if allow_reordering:
        order_sorting = np.argsort(proxy_order)
        proxy_order = proxy_order[order_sorting,]
        if ensure_nonnegative and proxy_order[0] < 0:
            return False
        new_order = [
            [order_sorting[i] if i >= 0 else i for i in row]
            for row in ordering
        ]
        return validate_zmatrix(new_order, allow_reordering=False)
    if ensure_nonnegative and proxy_order[0] < 0:
        # if raise_exception:
        #     raise ValueError("Z-matrix atom spec {} is non-zero")
        return False

    for n,row in enumerate(ordering):
        if (
                any(i > n for i in row)
                or any(i > row[0] for i in row[1:])
                or len(set(row)) < len(row)
        ):
            return False

    return True

def chain_zmatrix(n):
    return [
        list(range(i, i-4, -1))
        for i in range(n)
    ]

def center_bound_zmatrix(n, center=-1):
    return [
        [
            i,
            center,
            (
                (i - 2)
                if i > 1 else
                0
                if i == 1 else
                -2
            ),
            (
                (i - 1)
                if i > 1 else
                -3 + i
            ),
        ]
        for i in range(n)
    ]

def attached_zmatrix_fragment(n, fragment, attachment_points):
    return [
        [attachment_points[-r-1] if r < 0 else n+r for r in row]
        for row in fragment
    ]

def set_zmatrix_embedding(zmat, embedding=None):
    zmat = np.array(zmat)
    if embedding is None:
        embedding = [-1, -2, -3, -1, -2, -1]
    emb_pos = zmatrix_embedding_coords(zmat, array_inds=True)
    for (i,j),v in zip(emb_pos, embedding):
        zmat[..., i,j] = v
    return zmat

# ethyl_zmatrix = [
#     [0, -1, -2, -3],
#     [1,  0, -1, -2],
#     [2,  0,  1, -1]
# ]
#
# methyl_zmatrix = [
#     [0, -1, -2, -3],
#     [1,  0, -1, -2],
#     [2,  0,  1, -1],
#     [3,  0,  2,  1]
# ]


def functionalized_zmatrix(
        base_zm,
        attachments:'dict|list[list[int], list[int]]'=None,
        single_atoms:list[int]=None, # individual components, embedding doesn't matter
        methyl_positions:list[int]=None, # all bonds attached to central atom, angles relative to eachother
        ethyl_positions:list[int]=None, # all bonds attached to central atom, angles relative to eachother
):
    if nput.is_numeric(base_zm):
        zm = chain_zmatrix(base_zm)
    else:
        zm = [
            list(x) for x in base_zm
        ]
    if attachments is None:
        attachments = {}
    if hasattr(attachments, 'items'):
        attachments = attachments.items()
    for attachment_points, fragment in attachments:
        if nput.is_numeric(fragment):
            fragment = chain_zmatrix(fragment)
        zm = zm + attached_zmatrix_fragment(
            len(zm),
            fragment,
            attachment_points
        )
    if single_atoms is not None:
        for atom in single_atoms:
            zm = zm + attached_zmatrix_fragment(
                len(zm),
                [[0, -1, -2, -3]],
                [
                    (atom - i) if i < 0 else i
                    for i in range(atom, atom - 4, -1)
                ]
            )
    if methyl_positions is not None:
        for atom in methyl_positions:
            zm = zm + attached_zmatrix_fragment(
                len(zm),
                [
                    [0, -1, -2, -3],
                    [1, -1,  0, -2],
                    [2, -1,  0,  1],
                ],
                [
                    (atom - i) if i < 0 else i
                    for i in range(atom, atom - 4, -1)
                ]
            )
    if ethyl_positions is not None:
        for atom in ethyl_positions:
            zm = zm + attached_zmatrix_fragment(
                len(zm),
                [
                    [0, -1, -2, -3],
                    [1, -1,  0, -2]
                ],
                [
                    (atom - i) if i < 0 else i
                    for i in range(atom, atom - 4, -1)
                ]
            )
    return zm


def spoke_zmatrix(m, spoke=1, root=1):
    if nput.is_int(spoke):
        spoke = chain_zmatrix(spoke)

    if nput.is_int(root):
        root = chain_zmatrix(root)

    nroot = len(root) - 1
    if len(root) < 3:
        nrem = (3 - len(root))
        # no need for any moduli or floors, we just know
        if len(spoke) == 1:
            nspoke = nrem
        else:
            nspoke = 1

        for i in range(nspoke):
            if len(root) > 1:
                mroot = nroot + 1
            else:
                mroot = -1

            if len(root) > 2:
                proot = nroot + 2
            elif len(root) > 1:
                proot = nroot - 1
            else:
                proot = nroot - 2
            root = functionalized_zmatrix(
                root,
                [
                    [_attachment_point([
                        nroot,
                        mroot,
                        proot
                    ]), spoke]
                ]
            )
        if nspoke == 1:
            if nrem == 1:
                a = nroot - 1
                b = nroot + 1
            else:
                a = nroot + 1
                b = nroot + 2
        else:
            a = nroot + 1
            b = nroot + 2
        m = m - nspoke
    else:
        a = nroot - 1
        b = nroot - 2



    return functionalized_zmatrix(
        root,
        [
            [_attachment_point([nroot, a, b]), spoke]
            for _ in range(m)
        ]
    )


def reindex_zmatrix(zm, perm):
    return [
        [perm[r] if r >= 0 else r for r in row]
        for row in zm
    ]

def canonicalize_zmatrix(zm):
    if len(zm[0]) == 3:
        zm = [
            [0, -1, -2, -3]
        ] + [
            [i+1] + z
            for i,z in enumerate(zm)
        ]

    z_vec = np.array([z[0] for z in zm])
    perm = {z:i for i,z in enumerate(z_vec)}
    return z_vec, reindex_zmatrix(zm, perm)

def _attachment_point(i_pos):
    r = None
    a = None
    d = None
    if nput.is_int(i_pos):
        i_pos = [i_pos]

    if len(i_pos) > 0:
        r = i_pos[0]
    if len(i_pos) > 1:
        a = i_pos[1]
    if len(i_pos) > 2:
        d = i_pos[2]

    if r is None:
        if a is not None:
            if d is not None:
                if a > 0:
                    r = a - 1
                    if r == d: r = a + 1
                else:
                    r = a + 1
                    if r == d: r = a + 2
            else:
                r = (a - 1) if a > 0 else (a + 1)
        elif d is not None:
            if d > 1:
                r = d - 2
            elif d > 0:
                r = d - 1
            else:
                r = d + 1
        else:
            r = 0
    if a is None:
        if r > 0:
            a = r - 1
            if d is not None and a == d:
                if r > 1:
                    a = r - 2
                else:
                    a = r + 1
        else:
            a = r + 1
            if d is not None and a == d:
                a = r + 2
    if d is None:
        if r > 1:
            d = r - 2
            if a is not None and a == d:
                d = r - 1
        elif r > 0:
            d = r - 1
            if a is not None and a == d:
                d = r + 1
        else:
            d = r + 1
            if a is not None and a == d:
                d = r + 2

    return (r, a, d)
def add_missing_zmatrix_bonds(
        base_zmat,
        bonds,
        max_iterations=None
):
    atoms, zm = canonicalize_zmatrix(base_zmat)
    new_bonds = {}
    reindexing = list(atoms)
    for bi, be in bonds:
        if bi in atoms and be in atoms: continue
        if bi not in atoms and be not in atoms: continue
        if bi in atoms:
            # bi_pos = np.where(atoms == bi)[0][0]
            if bi not in new_bonds: new_bonds[bi] = []
            new_bonds[bi].append(be)
        if be in atoms:
            if be not in new_bonds: new_bonds[be] = []
            new_bonds[be].append(bi)

    if len(new_bonds) == 0:
        return base_zmat, new_bonds
    else:
        mods = {}
        for i,v in new_bonds.items():
            i_pos = np.where(atoms == i)[0][0]
            reindexing.extend(v)
            ix = _attachment_point(i_pos)
            mods[ix] = center_bound_zmatrix(len(v))

        new_zm = reindex_zmatrix(
            functionalized_zmatrix(
                zm,
                mods
            ),
            reindexing
        )

        if max_iterations is None or max_iterations > 0:
            new_zm, new_new_bonds = add_missing_zmatrix_bonds(
                new_zm,
                bonds,
                max_iterations=max_iterations-1 if max_iterations is not None else max_iterations
            )

            new_bonds.update(new_new_bonds)

        return new_zm, new_bonds


def bond_graph_zmatrix(
        bonds,
        fragments,
        edge_map=None,
        reindex=True
):
    submats = []
    backbone = fragments[0]
    primary = chain_zmatrix(len(backbone))
    if edge_map is None:
        edge_map = EdgeGraph.get_edge_map(bonds)
    for frag in fragments[1:]:
        if nput.is_int(frag[0]):
            submats.append(
                chain_zmatrix(len(frag))
            )
        else:
            submats.append(
                bond_graph_zmatrix(
                    bonds,
                    frag,
                    edge_map=edge_map,
                    reindex=False
                )
            )


    attachment_points = []
    for frag in fragments[1:]:
        if not nput.is_int(frag[0]):
            frag = frag[0]

        for f in frag:
            attach = None
            submap = edge_map.get(f)
            if nput.is_int(submap):
                if submap in backbone:
                    attach = submap
            else:
                for s in submap:
                    if s in backbone:
                        attach = s
                        break

            if attach is not None:
                attachment_points.append(backbone.index(attach))
                break

        else:
            raise ValueError("can't attach fragment to backbone, no connections")

    fused = functionalized_zmatrix(
        len(primary),
        {
            _attachment_point(ap):zmat
            for ap,zmat in zip(attachment_points, submats)
        }
    )

    if reindex:
        flat_frags = list(itut.flatten(fragments))
        fused = reindex_zmatrix(fused, flat_frags)

    return fused

def sort_complex_attachment_points(
        fragment_inds,
        attachment_points: 'dict|tuple[tuple[int], list[list[int]]]'
):
    new_attachments = [None] * len(fragment_inds)
    fragment_inds = list(fragment_inds)
    if hasattr(attachment_points, 'items'):
        attachment_points = attachment_points.items()
    for start, end in attachment_points:
        if nput.is_int(start):
            start = [start]
        if nput.is_int(end):
            end = [end]

        start_frag = None
        for i, f in enumerate(fragment_inds):
            if start[0] in f:
                start_frag = i
                break
        else:
            raise ValueError(f"index {start[0]} not in fragments {fragment_inds}")

        end_frag = None
        for i, f in enumerate(fragment_inds):
            if end[0] in f:
                end_frag = i
                break
        else:
            raise ValueError(f"index {start[0]} not in fragments {fragment_inds}")

        if end_frag < start_frag:
            start, end = end, start
            start_frag, end_frag = end_frag, start_frag
        elif end_frag == start_frag:
            raise ValueError(f"root index {start[0]} and end index {end[0]} in same fragment")

        fragment_inds[start_frag] = tuple(sorted(
            fragment_inds[start_frag],
            key=lambda i: start.index(i) if i in start else len(start)
        ))
        fragment_inds[end_frag] = tuple(sorted(
            fragment_inds[end_frag],
            key=lambda i: end.index(i) if i in end else len(end)
        ))

        new_attachments[end_frag] = start

    return fragment_inds, new_attachments

def complex_zmatrix(
        bonds,
        fragment_inds=None,
        fragment_zmats=None,
        distance_matrix=None,
        attachment_points=None,
        check_attachment_points=True,
        graph=None,
        reindex=True
):

    if fragment_inds is None:
        if fragment_zmats is not None:
            raise ValueError("can't supply just Z-mats, unclear which fragments they come from...")
        all_inds = np.unique(np.concatenate(bonds))
        if graph is None:
            graph = EdgeGraph(all_inds, bonds)

        fragment_inds = graph.get_fragments()

    all_inds = np.concatenate(fragment_inds)
    if graph is None:
        graph = EdgeGraph(all_inds, bonds)

    if fragment_zmats is None:
        if isinstance(attachment_points, dict):
            fragment_inds, attachment_points = sort_complex_attachment_points(
                fragment_inds,
                attachment_points
            )

        fragment_zmats = [
            bond_graph_zmatrix(bonds, f, edge_map=graph.map)
            for f in fragment_inds
        ]
    elif isinstance(attachment_points, dict) and check_attachment_points:
        raise ValueError("can't supply Z-mats and dict of attachment points, can't be sure attachments are right")

    inds = np.asanyarray(fragment_inds[0])
    zm = fragment_zmats[0]
    if attachment_points is None:
        attachment_points = [None] * len(fragment_inds)
    if len(attachment_points) < len(fragment_inds):
        raise ValueError("too few attachment points specified")
    for inds_2, zm_2, root in zip(fragment_inds[1:], fragment_zmats[1:], attachment_points[1:]):
        if root is None:
            if distance_matrix is None:
                subgraph = graph.take(inds)
                root = subgraph.get_centroid(check_fragments=False)
            else:
                distance_matrix = np.asanyarray(distance_matrix)
                dm = distance_matrix[np.ix_(inds, inds_2)]
                min_cols = np.argmin(dm, axis=1)
                min_row = np.argmin(dm[np.arange(len(inds)), min_cols], axis=0)
                root = inds[min_row]
                # root = np.where(inds == min_row)[0][0]
        else:
            if nput.is_int(root):
                root = np.where(inds == root)[0][0]
            else:
                root = [np.where(inds == r)[0][0] for r in root]

        inds = np.concatenate([inds, inds_2])
        zm = functionalized_zmatrix(
            zm,
            {
                _attachment_point(root): zm_2
            }
        )

    if reindex:
        zm = reindex_zmatrix(zm, inds)

    return zm



"""
Implements an importer for Gaussian output formats
"""

import numpy as np, re, math, io
from .GaussianLogComponents import GaussianLogComponents, GaussianLogDefaults, GaussianLogOrdering
from . import GaussianLogComponents as GaussianLogParsers
from .GaussianFChkComponents import FormattedCheckpointComponents, FormattedCheckpointCommonNames
from ...Parsers import FileStreamReader, FileStreamCheckPoint, FileStreamReaderException

__all__ = ["GaussianFChkReader", "GaussianLogReader", "GaussianLogReaderException", "GaussianFChkReaderException"]
__reload_hook__ = [ '.GaussianFChkComponents', ".GaussianLogComponents" ]

########################################################################################################################
#
#                                           GaussianLogReader
#
class GaussianLogReaderException(FileStreamReaderException):
    """
    A class for holding exceptions that occur in the course of reading from a log file
    """
    pass

class GaussianLogReader(FileStreamReader):
    """
    Implements a stream based reader for a Gaussian .log file.
    This is inherits from the `FileStreamReader` base, and takes a two pronged approach to getting data.
    First, a block is found in a log file based on a pair of tags.
    Next, a function (usually based on a `StringParser`) is applied to this data to convert it into a usable data format.
    The goal is to move toward wrapping all returned data in a `QuantityArray` so as to include data type information, too.

    You can see the full list of available keys in the `GaussianLogComponents` module, but currently they are:
    * `"Header"`: the header for the Gaussian job
    * `"InputZMatrix"`: the string of the input Z-matrix
    * `"CartesianCoordinates"`: all the Cartesian coordinates in the file
    * `"ZMatCartesianCoordinates"`: all of the Cartesian coordinate in Z-matrix orientation
    * `"StandardCartesianCoordinates"`: all of the Cartesian coordinates in 'standard' orientation
    * `"InputCartesianCoordinates"`: all of the Cartesian coordinates in 'input' orientation
    * `"ZMatrices"`: all of the Z-matrices
    * `"OptimizationParameters"`: all of the optimization parameters
    * `"MullikenCharges"`: all of the Mulliken charges
    * `"MultipoleMoments"`: all of the multipole moments
    * `"DipoleMoments"`: all of the dipole moments
    * `"OptimizedDipoleMoments"`: all of the dipole moments from an optimized scan
    * `"ScanEnergies"`: the potential surface information from a scan
    * `"OptimizedScanEnergies"`: the PES from an optimized scan
    * `"XMatrix"`: the anharmonic X-matrix from Gaussian's style of perturbation theory
    * `"Footer"`: the footer from a calculation

    You can add your own types, too.
    If you need something we don't have, give `GaussianLogComponents` a look to see how to add it in.

    """

    registered_components = GaussianLogComponents
    default_keys = GaussianLogDefaults
    default_ordering = GaussianLogOrdering
    parsers = GaussianLogParsers

    def parse(self, keys=None, num=None, reset=False):
        """The main function we'll actually use. Parses bits out of a .log file.

        :param keys: the keys we'd like to read from the log file
        :type keys: str or list(str)
        :param num: for keys with multiple entries, the number of entries to pull
        :type num: int or None
        :return: the data pulled from the log file, strung together as a `dict` and keyed by the _keys_
        :rtype: dict
        """
        if keys is None:
            keys = self.get_default_keys()
        # important for ensuring correctness of what we pull
        if isinstance(keys, str):
            keys = (keys,)
        keys = sorted(keys,
                      key = lambda k: (
                          -1 if (self.registered_components[k]["mode"] == "List") else (
                              self.default_ordering[k] if k in self.default_ordering else 0
                          )
                      )
                      )

        res = {}
        if reset:
            with FileStreamCheckPoint(self):
                for k in keys:
                    comp = self.registered_components[k]
                    res[k] = self.parse_key_block(**comp, num=num)
        else:
            for k in keys:
                comp = self.registered_components[k]
                try:
                    res[k] = self.parse_key_block(**comp, num=num)
                except:
                    raise GaussianLogReaderException("failed to parse block for key '{}'".format(k))
        return res

    job_default_keys = {
        "opt":{
            "p": ("StandardCartesianCoordinates", "OptimizedScanEnergies", "OptimizedDipoleMoments"),
            "_": ("StandardCartesianCoordinates", "OptimizedScanEnergies")
        },
        "popt": {
            "p": ("StandardCartesianCoordinates", "OptimizedScanEnergies", "OptimizedDipoleMoments"),
            "_": ("StandardCartesianCoordinates", "OptimizedScanEnergies")
        },
        "scan": ("StandardCartesianCoordinates", "ScanEnergies")
    }
    def get_default_keys(self):
        """
        Tries to get the default keys one might be expected to want depending on the type of job as determined from the Header
        Currently only supports 'opt', 'scan', and 'popt' as job types.

        :return: key listing
        :rtype: tuple(str)
        """
        header = self.parse("Header", reset=True)["Header"]

        header_low = {k.lower() for k in header.job}
        for k in self.job_default_keys:
            if k in header_low:
                sub = self.job_default_keys[k]
                if isinstance(sub, dict):
                    for k in sub:
                        if k in header_low:
                            defs = sub[k]
                            break
                    else:
                        defs = sub["_"]
                else:
                    defs = sub
                break
        else:
            raise GaussianLogReaderException("unclear what default keys should be used if not a scan and not a popt")

        return ("Header", ) + tuple(defs) + ("Footer",)

    @classmethod
    def read_props(cls, file, keys):
        with cls(file) as reader:
            parse = reader.parse(keys)
        if isinstance(keys, str):
            parse = parse[keys]
        return parse

########################################################################################################################
#
#                                           GaussianFChkReader
#
class GaussianFChkReaderException(FileStreamReaderException):
    pass

class GaussianFChkReader(FileStreamReader):
    """Implements a stream based reader for a Gaussian .fchk file. Pretty generall I think. Should be robust-ish.
    One place to change things up is convenient parsers for specific commonly pulled parts of the fchk

    """

    GaussianFChkReaderException = GaussianFChkReaderException
    registered_components = FormattedCheckpointComponents
    common_names = {to_:from_ for from_, to_ in FormattedCheckpointCommonNames.items()}
    to_common_name = FormattedCheckpointCommonNames

    def __init__(self, file, **kwargs):
        super().__init__(file, **kwargs)
        self._num_atoms = None
        # with self: self.num_atoms = self.parse("Number of atoms")["Number of atoms"]

    def read_header(self):
        """Reads the header and skips the stream to where we want to be

        :return: the header
        :rtype: str
        """
        return self.get_tagged_block(None, "Number of atoms")

    fchk_re_pattern = r"^(.+?)\s+(I|R|C|H)\s+(N=)?\s+(.+)\s+" # matches name, type, num (if there), and val
    fchk_re = re.compile(fchk_re_pattern)
    def get_next_block_params(self):
        """Pulls the tag of the next block, the type, the number of bytes it'll be,
        and if it's a single-line block it'll also spit back the block itself

        :return:
        :rtype: dict
        """
        with FileStreamCheckPoint(self):
            tag_line = self.readline()
            if tag_line == b'' or tag_line == '':
                return None
            match = re.match(self.fchk_re, tag_line)
            if match is None:
                with FileStreamCheckPoint(self):
                    for i in range(4):
                        prev_lines = self.rfind("\n")
                        self.seek(prev_lines)
                    lines = "".join(self.readline() for i in range(4))
                    raise GaussianFChkReaderException("{}.{}: line '{}' couldn't be read as a tag line (in '{}')".format(
                        type(self).__name__,
                        "get_next_block_params",
                        tag_line,
                        lines
                    ))
            jump = self.tell()
        self.seek(jump)

        name, btype, numQ, val = gg = match.groups()
        # print(gg)
        if numQ:
            byte_count = 0
            shits = int(val)
            # hard coded these block formats since they're documented by Gaussian and thus unlikely to change
            if btype == "I":
                byte_per_shit = 12
                shits_per_line = 6
                btype = int
            elif btype == "R":
                byte_per_shit = 16
                shits_per_line = 5
                btype = float
            elif btype == "C":
                byte_per_shit = 12
                shits_per_line = 5
                btype = str
            elif btype == "L":
                byte_per_shit = 12
                shits_per_line = 5
                btype = bool
            byte_count = shits * byte_per_shit + math.ceil( shits / shits_per_line ) # each newline needs a byte
        else:
            byte_count = None
            if btype == "I":
                val = int(val)
            elif btype == "R":
                val = float(val)
            elif btype == "L":
                val = bool(int(val))

        return {
            "name": name,
            "dtype": btype,
            "byte_count": byte_count,
            "value": val
        }

    def get_block(self, name = None, dtype = None, byte_count = None, value = None):
        """Pulls the next block by first pulling the block tag

        :return:
        :rtype:
        """

        if byte_count is not None:

            block_str = self.read(byte_count)
            if dtype in {int, float}:
                block_str = io.StringIO(block_str.replace("\n", "")) # flatten it out
                value = np.loadtxt(block_str)
                if dtype == int:
                    value = value.astype(np.int64)
            else:
                value = block_str

        # try:
        parser = self.registered_components.get(name, None)
        if parser is not None:
            value = parser(value, reader=self)
        # except KeyError:
        #     pass

        return value

    def skip_block(self, name = None, dtype = None, byte_count = None, value = None):
        """Skips the next block

        :return:
        :rtype:
        """

        if byte_count is not None:
            self.seek(self.tell() + byte_count)

    @property
    def num_atoms(self):
        if self._num_atoms is None:
            self._num_atoms = self.parse(["Number of atoms"])["Number of atoms"]
        return self._num_atoms
    def parse(self, keys=None, default='raise'):
        if keys is None:
            keys_to_go = None
        else:
            if isinstance(keys, str):
                keys = (keys,)
            keys_original = set(keys)
            keys_to_go = { (self.common_names[k] if k in self.common_names else k) for k in keys }

        parse_results = {}
        header = self.read_header()
        if self._num_atoms is None:
            self._num_atoms = self.get_block(**self.get_next_block_params())
        parse_results["Number of atoms"] = self._num_atoms

        parse_results['Header'] = header
        if keys_to_go is None:
            while True: # I'll just break once I've exhausted everything
                next_block = self.get_next_block_params()
                if next_block is None:
                    break
                tag = next_block["name"]
                parse_results[tag] = self.get_block(**next_block)
        else:
            while len(keys_to_go)>0:
                # try to skip malformatted blocks...
                try:
                    next_block = self.get_next_block_params()
                except GaussianFChkReaderException:
                    fp = self.find("\n")
                    if fp == -1:
                        next_block = None
                    else:
                        next_block = ""
                        self.seek(fp + 1)
                    while next_block is not None and next_block == "":
                        try:
                            next_block = self.get_next_block_params()
                        except GaussianFChkReaderException:
                            fp = self.find("\n")
                            if fp == -1:
                                next_block = None
                            else:
                                self.seek(fp + 1)

                if next_block is None:
                    if isinstance(default, str) and default == 'raise':
                        raise GaussianFChkReaderException("{}.{}: couldn't find keys {}".format(
                            type(self).__name__,
                            "parse",
                            keys_to_go
                            )
                        )
                    else:
                        for tag in keys_to_go:
                            if tag not in keys_original:
                                tag = self.to_common_name[tag]
                            parse_results[tag] = default
                        break
                tag = next_block["name"]
                if tag in keys_to_go:
                    keys_to_go.remove(tag)
                    if tag not in keys_original:
                        tag = self.to_common_name[tag]
                    parse_results[tag] = self.get_block(**next_block)
                else:
                    self.skip_block(**next_block)

        return parse_results

    @classmethod
    def read_props(cls, file, keys):
        with cls(file) as reader:
            parse = reader.parse(keys)
        if isinstance(keys, str):
            parse = parse[keys]
        return parse





import os
from .WebAPI import *

__all__ = [
    "ChemSpiderAPI"
]

class ChemSpiderAPI(WebAPIConnection):
    """
    It is better in general to just use the ChemSpiderPy package, but this works for now
    """

    request_base = 'https://api.rsc.org/compounds/v1'
    api_key_env_var = "CHEM_SPIDER_APIKEY"
    def __init__(self, token=None, request_delay_time=None,  **opts):
        token = self.get_chemspider_apikey(token)
        super().__init__({'header':'apikey', 'value':token}, request_delay_time=request_delay_time, **opts)

    @classmethod
    def get_chemspider_apikey(cls, token):
        if token is None:
            token = os.environ.get(cls.api_key_env_var)
        return token

    @property
    def filter(self):
        return self.get_subapi('filter')

    @property
    def records(self):
        return self.get_subapi('records')

    @property
    def lookups(self):
        return self.get_subapi('lookups')

    @property
    def tool(self):
        return self.get_subapi('tool')

    def handle_filter_query(self, query_id, count=1, start=0, **polling_opts):
        res = self.filter.get(f'{query_id}/results', query=dict(count=count, start=start))
        return res
        # https: // api.rsc.org / compounds / v1
        # / filter / {queryId} / results

    def apply_filter_query(self, filter_path, retries=None, timeout=None, request_delay_time=None, **opts):
        base_query = self.filter.post(filter_path, json=opts)
        if 'queryId' in base_query:
            # TODO: handle status?
            request_delay_time = request_delay_time
            return self.handle_filter_query(base_query['queryId'], retries=retries, timeout=timeout)
        else:
            raise ValueError("`queryId` not in filter JSON")

    default_molecule_fields = [
        "CommonName",
        "SMILES",
        "InChI"
    ]
    def get_info(self, ids, fields=None, **opts):
        if fields is None:
            fields = self.default_molecule_fields
        elif isinstance(fields, str):
            fields = [fields]
        return self.records.post("batch", json=dict(recordIds=ids, fields=fields), **opts)['records']

    _name_cache = {}
    def get_compounds_by_name(self, name, return_ids=False, fields=None, **opts):
        if name not in self._name_cache:
            self._name_cache[name] = self.apply_filter_query('name', name=name, **opts)
        ids = self._name_cache[name]['results']
        if return_ids:
            return ids
        else:
            return self.get_info(ids, fields=fields)


        # payload = {"name": "aspirin",
        #            "orderBy": "default",
        #            "orderDirection": "default"}
        #
        # headers = {"apikey": "*your_API_Key*",
        #            "Content-Type": "application/json",
        #            "Accept": "application/json"}
        #
        # response = requests.post(url, json=payload, headers=headers)
        # print(response.json())
        # )